# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checks needed to determine integrity of workflows."""
from datetime import timedelta
from typing import List, Optional, Tuple, cast

from renku.command.command_builder import inject
from renku.command.util import WARNING
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.util import communication
from renku.domain_model.project_context import project_context
from renku.domain_model.workflow.plan import AbstractPlan
from renku.infrastructure.gateway.activity_gateway import reindex_catalog


def check_activity_catalog(fix, force, **_) -> Tuple[bool, bool, Optional[str]]:
    """Check if the activity-catalog needs to be rebuilt.

    Args:
        fix: Whether to fix found issues.
        force: Whether to force rebuild the activity catalog.
        _: keyword arguments.

    Returns:
        Tuple of whether the activity-catalog needs to be rebuilt, if an automated fix is available and a string of
            found problems.
    """
    database = project_context.database
    activity_catalog = database["activity-catalog"]
    relations = database["_downstream_relations"]

    # NOTE: If len(activity_catalog) > 0 then either the project is fixed or it used a fixed Renku version but still has
    # broken metadata. ``force`` allows to rebuild the metadata in the latter case.
    if (len(relations) == 0 or len(activity_catalog) > 0) and not (force and fix):
        return True, False, None

    if not fix:
        problems = (
            WARNING + "The project's workflow metadata needs to be rebuilt (use 'renku doctor --fix' to rebuild it).\n"
        )

        return False, True, problems

    with communication.busy("Rebuilding workflow metadata ..."):
        reindex_catalog(database=database)

    communication.info("Workflow metadata was rebuilt")

    return True, False, None


@inject.autoparams("plan_gateway")
def check_plan_modification_date(fix, plan_gateway: IPlanGateway, **_) -> Tuple[bool, bool, Optional[str]]:
    """Check if all plans have modification date set for them.

    Args:
        fix(bool): Whether to fix found issues.
        plan_gateway(IPlanGateway): Injected PlanGateway.
        _: keyword arguments.

    Returns:
        Tuple[bool, Optional[str]]: Tuple of whether there are plans without modification date, if an automated fix is
            available and a string of their IDs
    """
    plans: List[AbstractPlan] = plan_gateway.get_all_plans()

    to_be_processed = []
    for plan in plans:
        if not hasattr(plan, "date_modified") or plan.date_modified is None:
            to_be_processed.append(plan)

    if not to_be_processed:
        return True, False, None
    if not fix:
        ids = [plan.id for plan in to_be_processed]
        message = (
            WARNING
            + "The following workflows have incorrect modification date (use 'renku doctor --fix' to fix them):\n\t"
            + "\n\t".join(ids)
        )
        return False, True, message

    fix_plan_dates(plans=to_be_processed, plan_gateway=plan_gateway)
    project_context.database.commit()
    communication.info("Workflow modification dates were fixed")

    return True, False, None


def fix_plan_dates(plans: List[AbstractPlan], plan_gateway):
    """Set modification date on a list of plans and fix their creation date."""
    processed = set()
    # NOTE: switch creation date for modification date
    for tail in plans:
        to_be_processed: List[AbstractPlan] = []
        if tail not in processed:
            processed.add(tail)
            to_be_processed.append(tail)
        creation_date = tail.date_created
        plan = tail

        while plan.is_derivation():
            plan = cast(AbstractPlan, plan_gateway.get_by_id(plan.derived_from))
            creation_date = plan.date_created
            if plan not in processed:
                processed.add(plan)
                to_be_processed.append(plan)

        while to_be_processed:
            plan = to_be_processed.pop()
            plan.unfreeze()
            plan.date_modified = plan.date_created
            plan.date_created = creation_date
            if plan.date_removed and plan.date_removed < plan.date_created:
                plan.date_removed = plan.date_created + timedelta(seconds=1)
            plan.freeze()


@inject.autoparams("plan_gateway")
def check_plan_id(fix, plan_gateway: IPlanGateway, **_) -> Tuple[bool, bool, Optional[str]]:
    """Check if all plans have correct IDs.

    Args:
        fix(bool): Whether to fix found issues.
        plan_gateway(IPlanGateway): Injected PlanGateway.
        _: keyword arguments.

    Returns:
        Tuple[bool, Optional[str]]: Tuple of whether there are plans with invalid IDs, if an automated fix is
            available and a string of their IDs
    """
    plans: List[AbstractPlan] = plan_gateway.get_all_plans()

    to_be_processed = []
    for plan in plans:
        if isinstance(plan.id, str) and plan.id.startswith("/plans//plans"):
            to_be_processed.append(plan)

    if not to_be_processed:
        return True, False, None
    if not fix:
        ids = [plan.id for plan in to_be_processed]
        message = (
            WARNING
            + "The following workflows have incorrect IDs (use 'renku doctor --fix' to fix them):\n\t"
            + "\n\t".join(ids)
        )
        return False, True, message

    for plan in to_be_processed:
        plan.unfreeze()
        plan.id = plan.id.replace("//plans/", "/")
        plan.freeze()
    project_context.database.commit()
    communication.info("Workflow IDs were fixed")

    return True, False, None



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku graph export controller."""

from requests import RequestException
from sentry_sdk import capture_exception

from renku.command.graph import export_graph_command
from renku.command.migrate import migrations_check
from renku.command.view_model.graph import DotFormat
from renku.core.errors import RenkuException
from renku.ui.service.config import PROJECT_CLONE_NO_DEPTH
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.serializers.graph import (
    GraphExportCallbackError,
    GraphExportCallbackSuccess,
    GraphExportRequest,
    GraphExportResponseRPC,
)
from renku.ui.service.views import result_response


class GraphExportCtrl(ServiceCtrl, RenkuOperationMixin):
    """Controller for export graph endpoint."""

    REQUEST_SERIALIZER = GraphExportRequest()
    RESPONSE_SERIALIZER = GraphExportResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a datasets list controller."""
        self.ctx = GraphExportCtrl.REQUEST_SERIALIZER.load(request_data)
        super().__init__(cache, user_data, request_data, clone_depth=PROJECT_CLONE_NO_DEPTH)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        result = migrations_check().build().execute().output

        if not result.project_supported:
            raise RenkuException("project not supported")

        callback_payload = {
            "project_url": self.context["git_url"],
            "commit_id": self.context["revision"] or "master",
        }

        try:
            result = export_graph_command().build().execute(revision_or_range=self.context["revision"])

            format = self.context["format"]

            if format == "json-ld":
                result = result.output.as_jsonld_string(indentation=None)
            elif format == "rdf":
                result = result.output.as_rdf_string()
            elif format == "nt":
                result = result.output.as_nt_string()
            elif format == "dot":
                result = result.output.as_dot_string(format=DotFormat.FULL)
            elif format == "dot-landscape":
                result = result.output.as_dot_string(format=DotFormat.FULL_LANDSCAPE)
            else:
                raise NotImplementedError(f"Format {format} is not supported on this endpoint.")

            if self.context.get("callback_url"):
                self.report_success(callback_payload, {"payload": result}, self.context["callback_url"])

            return result
        except (RequestException, RenkuException, MemoryError) as e:
            if self.context.get("callback_url"):
                self.report_recoverable(callback_payload, e, self.context["callback_url"])
            raise
        except BaseException as e:
            if self.context.get("callback_url"):
                self.report_unrecoverable(callback_payload, e, self.context["callback_url"])
            raise

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        self.ctx["graph"] = self.execute_op()
        return result_response(GraphExportCtrl.RESPONSE_SERIALIZER, self.ctx)

    def report_recoverable(self, payload, exception, callback_url):
        """Report to callback URL recoverable state."""
        from renku.core.util import requests

        capture_exception(exception)

        if not callback_url:
            return

        payload["failure"] = {"type": "RECOVERABLE_FAILURE", "message": str(exception)}

        data = GraphExportCallbackError().load(payload)
        requests.post(callback_url, data=data)

    def report_unrecoverable(self, payload, exception, callback_url):
        """Report to callback URL unrecoverable state."""
        from renku.core.util import requests

        capture_exception(exception)

        if not callback_url:
            return

        payload["failure"] = {"type": "UNRECOVERABLE_FAILURE", "message": str(exception)}

        data = GraphExportCallbackError().load(payload)
        requests.post(callback_url, data=data)

    def report_success(self, request_payload, graph_payload, callback_url):
        """Report to callback URL success state."""
        from renku.core.util import requests

        data = GraphExportCallbackSuccess().load({**request_payload, **graph_payload})

        if not callback_url:
            return data

        requests.post(callback_url, data=data)

        return data



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent an execution of a Plan."""

from datetime import datetime
from itertools import chain
from typing import Dict, List, Optional, Union, cast
from uuid import uuid4

import deal
from werkzeug.utils import cached_property

from renku.command.command_builder import inject
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.util.datetime8601 import local_now
from renku.core.util.git import get_entity_from_revision, get_git_user
from renku.domain_model.entity import Collection, Entity
from renku.domain_model.provenance.agent import Person, SoftwareAgent
from renku.domain_model.provenance.annotation import Annotation
from renku.domain_model.provenance.parameter import ParameterValue
from renku.domain_model.workflow.plan import Plan
from renku.domain_model.workflow.workflow_file import WorkflowFileCompositePlan
from renku.infrastructure.database import Persistent
from renku.infrastructure.immutable import Immutable
from renku.infrastructure.repository import Repository
from renku.version import __version__, version_url


class Association:
    """Assign responsibility to an agent for an activity."""

    def __init__(self, *, agent: Union[Person, SoftwareAgent], id: str, plan: Union[Plan, WorkflowFileCompositePlan]):
        self.agent: Union[Person, SoftwareAgent] = agent
        self.id: str = id
        self.plan: Union[Plan, WorkflowFileCompositePlan] = plan

    @staticmethod
    def generate_id(activity_id: str) -> str:
        """Generate a Association identifier."""
        return f"{activity_id}/association"


class Usage(Immutable):
    """Represent a dependent path."""

    __slots__ = ("entity",)

    entity: Union[Collection, Entity]
    id: str

    def __init__(self, *, entity: Union[Collection, Entity], id: str):
        super().__init__(entity=entity, id=id)

    @staticmethod
    def generate_id(activity_id: str) -> str:
        """Generate a Usage identifier."""
        return f"{activity_id}/usages/{uuid4().hex}"


class HiddenUsage(Usage):
    """Represent a dependent path corresponding to a ``HiddenInput``."""


class Generation(Immutable):
    """Represent an act of generating a path."""

    __slots__ = ("entity",)

    entity: Union[Collection, Entity]
    id: str

    def __init__(self, *, entity: Union[Collection, Entity], id: str):
        super().__init__(entity=entity, id=id)

    @staticmethod
    def generate_id(activity_id: str) -> str:
        """Generate a Generation identifier."""
        return f"{activity_id}/generations/{uuid4().hex}"


class Activity(Persistent):
    """Represent an activity in the repository."""

    invalidated_at: Optional[datetime] = None
    hidden_usages: List[HiddenUsage] = list()

    @deal.ensure(lambda self, *_, result, **kwargs: self.ended_at_time >= self.started_at_time)
    def __init__(
        self,
        *,
        agents: List[Union[Person, SoftwareAgent]],
        annotations: Optional[List[Annotation]] = None,
        association: Association,
        ended_at_time: datetime,
        generations: Optional[List[Generation]] = None,
        hidden_usages: Optional[List[HiddenUsage]] = None,
        id: str,
        invalidated_at: Optional[datetime] = None,
        invalidations: Optional[List[Entity]] = None,
        parameters: Optional[List[ParameterValue]] = None,
        project_id: Optional[str] = None,
        started_at_time: datetime,
        usages: Optional[List[Usage]] = None,
    ):
        self.agents: List[Union[Person, SoftwareAgent]] = agents
        self.annotations: List[Annotation] = annotations or []
        self.association: Association = association
        self.ended_at_time: datetime = ended_at_time
        self.generations: List[Generation] = generations or []
        self.hidden_usages: List[HiddenUsage] = hidden_usages or []
        self.id: str = id
        self.invalidated_at: Optional[datetime] = invalidated_at
        self.invalidations: List[Entity] = invalidations or []
        self.parameters: List[ParameterValue] = parameters or []
        self.project_id: Optional[str] = project_id
        self.started_at_time: datetime = started_at_time
        self.usages: List[Usage] = usages or []

        if not self.id.startswith("/activities/"):
            self.id = f"/activities/{self.id}"

        # TODO: _was_informed_by = attr.ib(kw_only=True)
        # TODO: influenced = attr.ib(kw_only=True)

    @classmethod
    @deal.post(lambda activity: activity is None or activity.ended_at_time >= activity.started_at_time)
    # NOTE: check for none due to bug in deal thinking inner function return value is actual return
    @inject.autoparams("project_gateway")
    def from_plan(
        cls,
        plan: Plan,
        repository: "Repository",
        project_gateway: IProjectGateway,
        started_at_time: datetime,
        ended_at_time: datetime,
        annotations: Optional[List[Annotation]] = None,
        id: Optional[str] = None,
    ):
        """Convert a ``Plan`` to a ``Activity``."""
        from renku.core.plugin.pluginmanager import get_plugin_manager

        usages: Dict[str, Usage] = {}
        hidden_usages: Dict[str, HiddenUsage] = {}
        generations = {}
        parameter_values = []

        activity_id = id or cls.generate_id()

        def process_input(input, already_processed, cls, add_parameter_value):
            input_path = input.actual_value

            if add_parameter_value:
                parameter_values.append(
                    ParameterValue(id=ParameterValue.generate_id(activity_id), parameter_id=input.id, value=input_path)
                )

            if input_path in already_processed:
                return

            entity = get_entity_from_revision(repository=repository, path=input_path, bypass_cache=True)

            dependency = cls(entity=entity, id=cls.generate_id(activity_id))

            already_processed[input_path] = dependency

        for input in plan.inputs:
            process_input(input=input, already_processed=usages, cls=Usage, add_parameter_value=True)

        for input in plan.hidden_inputs:
            process_input(input=input, already_processed=hidden_usages, cls=HiddenUsage, add_parameter_value=False)

        for output in plan.outputs:
            output_path = output.actual_value

            parameter_values.append(
                ParameterValue(id=ParameterValue.generate_id(activity_id), parameter_id=output.id, value=output_path)
            )

            if output_path in generations:
                continue

            entity = get_entity_from_revision(repository=repository, path=output_path, bypass_cache=True)

            generation = Generation(entity=entity, id=Generation.generate_id(activity_id))

            generations[output_path] = generation

        for parameter in plan.parameters:
            value = parameter.actual_value

            parameter_values.append(
                ParameterValue(id=ParameterValue.generate_id(activity_id), parameter_id=parameter.id, value=value)
            )

        agent = SoftwareAgent(id=version_url, name=f"renku {__version__}")
        person = cast(Person, get_git_user(repository))
        association = Association(agent=agent, id=Association.generate_id(activity_id), plan=plan)

        activity = cls(
            id=activity_id,
            association=association,
            agents=[agent, person],
            usages=list(usages.values()),
            hidden_usages=list(hidden_usages.values()),
            generations=list(generations.values()),
            parameters=parameter_values,
            project_id=project_gateway.get_project().id,
            started_at_time=started_at_time,
            ended_at_time=ended_at_time,
            annotations=annotations,
        )

        pm = get_plugin_manager()

        plugin_annotations = list(chain.from_iterable(pm.hook.activity_annotations(activity=activity)))

        if plugin_annotations:
            activity.annotations.extend(plugin_annotations)

        return activity

    def __repr__(self):
        return f"<Activity '{self.id}': {self.association.plan.name} @ {self.ended_at_time}>"

    @cached_property
    def plan_with_values(self) -> Plan:
        """Get a copy of the associated plan with values from ParameterValues applied."""
        plan = self.association.plan.copy()

        for parameter in self.parameters:
            parameter.apply_value_to_parameter(plan)

        return plan

    @property
    def deleted(self) -> bool:
        """Return if the activity was deleted."""
        return self.invalidated_at is not None

    @staticmethod
    def generate_id(uuid: Optional[str] = None) -> str:
        """Generate an identifier for an activity."""
        if uuid is None:
            uuid = uuid4().hex
        return f"/activities/{uuid}"

    def has_identical_inputs_and_outputs_as(self, other: "Activity"):
        """Return true if all input and outputs paths are identical regardless of the order."""
        return sorted(u.entity.path for u in self.usages) == sorted(u.entity.path for u in other.usages) and sorted(
            g.entity.path for g in self.generations
        ) == sorted(g.entity.path for g in other.generations)

    def compare_to(self, other: "Activity") -> int:
        """Compare execution date with another activity; return a positive value if self is executed after the other."""
        if self.ended_at_time < other.ended_at_time:
            return -1
        elif self.ended_at_time > other.ended_at_time:
            return 1
        elif self.started_at_time < other.started_at_time:
            return -1
        elif self.started_at_time > other.started_at_time:
            return 1

        return 0

    @deal.ensure(lambda self, *_, result, **kwargs: self.invalidated_at >= self.ended_at_time)
    def delete(self, when: Optional[datetime] = None):
        """Mark the activity as deleted."""
        self.unfreeze()
        self.invalidated_at = when or local_now()
        self.freeze()


class ActivityCollection(Persistent):
    """Represent a list of activities."""

    def __init__(self, *, activities: List[Activity], id: Optional[str] = None):
        self.activities: List[Activity] = activities or []
        self.id: str = id or self.generate_id()

    @staticmethod
    def generate_id() -> str:
        """Generate an identifier for an activity collection."""
        return f"/activity-collection/{uuid4().hex}"


class WorkflowFileActivityCollection(ActivityCollection):
    """Represent activities of a workflow file execution."""

    def __init__(
        self,
        *,
        activities: List[Activity],
        agents: List[Union[Person, SoftwareAgent]],
        association: Association,
        ended_at_time: datetime,
        id: Optional[str] = None,
        invalidated_at: Optional[datetime] = None,
        project_id: Optional[str] = None,
        started_at_time: datetime,
    ):
        super().__init__(activities=activities, id=id)

        self.agents: List[Union[Person, SoftwareAgent]] = agents
        self.association: Association = association
        self.ended_at_time: datetime = ended_at_time
        self.invalidated_at: Optional[datetime] = invalidated_at
        self.project_id: Optional[str] = project_id
        self.started_at_time: datetime = started_at_time

    @classmethod
    @inject.autoparams("project_gateway")
    def from_activities(
        cls, plan: WorkflowFileCompositePlan, project_gateway: IProjectGateway, activities: List[Activity]
    ):
        """Create an instance from a list of ``Activity``."""
        id = cls.generate_id()
        association = Association(agent=activities[0].association.agent, id=Association.generate_id(id), plan=plan)

        return cls(
            activities=activities,
            agents=activities[0].agents.copy(),
            association=association,
            ended_at_time=max(a.ended_at_time for a in activities),
            id=id,
            invalidated_at=None,
            project_id=project_gateway.get_project().id,
            started_at_time=min(a.started_at_time for a in activities),
        )

    @staticmethod
    def generate_id() -> str:
        """Generate an identifier."""
        return f"/workflow-file-activity-collection/{uuid4().hex}"

    def __repr__(self):
        return f"<WorkflowFileActivityCollection '{self.id}': {self.association.plan.name} @ {self.ended_at_time}>"



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Web dataset provider."""

import urllib
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

from renku.core import errors
from renku.core.constant import CACHE
from renku.core.dataset.dataset_add import copy_file
from renku.core.dataset.providers.api import AddProviderInterface, ProviderApi, ProviderPriority
from renku.core.util import communication
from renku.core.util.os import delete_dataset_file
from renku.core.util.urls import check_url, remove_credentials
from renku.core.util.util import parallel_execute
from renku.domain_model.project_context import project_context
from renku.infrastructure.immutable import DynamicProxy

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata, DatasetUpdateMetadata


class WebProvider(ProviderApi, AddProviderInterface):
    """A provider for downloading data from web URLs."""

    priority = ProviderPriority.LOWEST
    name = "Web"
    is_remote = True

    @staticmethod
    def supports(uri: str) -> bool:
        """Whether or not this provider supports a given URI."""
        is_remote, is_git = check_url(uri)
        return is_remote and not is_git

    def get_metadata(
        self,
        uri: str,
        destination: Path,
        *,
        extract: bool = False,
        filename: Optional[str] = None,
        multiple: bool = False,
        **kwargs,
    ) -> List["DatasetAddMetadata"]:
        """Get metadata of files that will be added to a dataset."""
        dataset = kwargs.get("dataset")
        if dataset and dataset.storage and urlparse(dataset.storage).scheme != urlparse(uri).scheme:
            raise errors.ParameterError(
                f"The scheme of the url {uri} does not match the defined storage url {dataset.storage}."
            )

        return download_file(
            project_path=project_context.path,
            uri=uri,
            destination=destination,
            extract=extract,
            filename=filename,
            multiple=multiple,
        )

    def update_files(
        self,
        files: List[DynamicProxy],
        dry_run: bool,
        delete: bool,
        context: Dict[str, Any],
        **kwargs,
    ) -> List["DatasetUpdateMetadata"]:
        """Update dataset files from the remote provider."""
        from renku.core.dataset.providers.models import DatasetAddMetadata, DatasetUpdateAction, DatasetUpdateMetadata

        progress_text = "Checking for local updates"
        results: List[DatasetUpdateMetadata] = []

        download_cache: Dict[str, DatasetAddMetadata] = {}
        potential_updates: List[Tuple[DatasetAddMetadata, DynamicProxy]] = []

        try:
            communication.start_progress(progress_text, len(files))
            for file in files:
                if not file.source:
                    continue
                destination = project_context.path / file.dataset.get_datadir()
                try:
                    if file.entity.path not in download_cache:
                        downloaded_files = download_file(
                            project_path=project_context.path, uri=file.source, destination=destination
                        )

                        if not any(f.entity_path == file.entity.path for f in downloaded_files):
                            # File probably comes from an extracted download
                            downloaded_files = download_file(
                                project_path=project_context.path,
                                uri=file.source,
                                destination=destination,
                                extract=True,
                            )

                        download_cache.update({str(f.entity_path): f for f in downloaded_files})
                except errors.OperationError:
                    results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.DELETE))
                else:
                    metadata = download_cache.get(file.entity.path)

                    if not metadata:
                        results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.DELETE))

                        if not dry_run and delete:
                            delete_dataset_file(file.entity.path, follow_symlinks=True)
                            project_context.repository.add(file.entity.path, force=True)
                    else:
                        potential_updates.append((metadata, file))

        finally:
            communication.finalize_progress(progress_text)

        if not potential_updates:
            return results

        check_paths: List[Union[Path, str]] = [
            str(u[0].source.relative_to(project_context.path)) for u in potential_updates
        ]
        # Stage files temporarily so we can get hashes
        project_context.repository.add(*check_paths, force=True)
        hashes = project_context.repository.get_object_hashes(check_paths)
        project_context.repository.remove(*check_paths, index=True)

        for metadata, file in potential_updates:
            if file.entity.checksum != hashes.get(metadata.source):
                results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.UPDATE))
                if not dry_run:
                    copy_file(metadata, file.dataset, storage=None)
        return results


def _ensure_dropbox(url):
    """Ensure dropbox url is set for file download."""
    if not isinstance(url, urllib.parse.ParseResult):
        url = urllib.parse.urlparse(url)

    query = url.query or ""
    if "dl=0" in url.query:
        query = query.replace("dl=0", "dl=1")
    else:
        query += "dl=1"

    url = url._replace(query=query)
    return url


def _provider_check(url):
    """Check additional provider related operations."""
    from renku.core.util import requests

    url = requests.get_redirect_url(url)
    url = urllib.parse.urlparse(url)

    if "dropbox.com" in url.netloc:
        url = _ensure_dropbox(url)

    return urllib.parse.urlunparse(url)


def download_file(
    uri: str,
    filename: Optional[str] = None,
    *,
    project_path: Path,
    destination: Path,
    extract: bool = False,
    multiple: bool = False,
) -> List["DatasetAddMetadata"]:
    """Download a file from a URI and return its metadata."""
    from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata
    from renku.core.util import requests

    uri = requests.get_redirect_url(uri)  # TODO: Check that this is not duplicate
    uri = _provider_check(uri)

    with project_context.with_path(project_path):
        try:
            # NOTE: If execution time was less than the delay, block the request until delay seconds are passed
            tmp_root, paths = requests.download_file(
                base_directory=project_context.metadata_path / CACHE, url=uri, filename=filename, extract=extract
            )
        except errors.RequestError as e:  # pragma nocover
            raise errors.OperationError(f"Cannot download from {uri}") from e

        paths = [p for p in paths if not p.is_dir()]

        if len(paths) > 1 or multiple:
            if destination.exists() and not destination.is_dir():
                raise errors.ParameterError(f"Destination is not a directory: '{destination}'")
            destination.mkdir(parents=True, exist_ok=True)
        elif len(paths) == 1:
            tmp_root = paths[0].parent if destination.exists() else paths[0]

        paths = [(src, destination / src.relative_to(tmp_root)) for src in paths if not src.is_dir()]

        return [
            DatasetAddMetadata(
                entity_path=dst.relative_to(project_context.path),
                url=remove_credentials(uri),
                action=DatasetAddAction.MOVE,
                source=src,
                destination=dst,
            )
            for src, dst in paths
        ]


def download_files(
    urls: Tuple[str, ...], destination: Path, names: Tuple[str, ...], extract: bool
) -> List["DatasetAddMetadata"]:
    """Download multiple files and return their metadata."""
    assert len(urls) == len(names), f"Number of URL and names don't match {len(urls)} != {len(names)}"

    if destination.exists() and not destination.is_dir():
        raise errors.ParameterError(f"Destination is not a directory: '{destination}'")

    destination.mkdir(parents=True, exist_ok=True)

    return parallel_execute(
        download_file,
        urls,
        names,
        project_path=project_context.path,
        destination=destination,
        extract=extract,
        multiple=True,
    )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for tracking parameter values in provenance."""

from typing import Any
from uuid import uuid4

from renku.domain_model.workflow.plan import Plan


class ParameterValue:
    """Value for a parameter in provenance."""

    def __init__(self, *, id: str, parameter_id: str, value: Any):
        self.id = id
        self.parameter_id: str = parameter_id
        self.value: Any = value

    @staticmethod
    def generate_id(activity_id: str) -> str:
        """Generate a default id."""
        return f"{activity_id}/parameter-value/{uuid4().hex}"

    def apply_value_to_parameter(self, plan: Plan) -> None:
        """Apply the current value as actual_value on the plan's parameter."""
        parameter = plan.get_field_by_id(self.parameter_id)
        parameter.actual_value = self.value



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service jobs views."""

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.errors import IntermittentProjectIdError
from renku.ui.service.serializers.jobs import JobDetailsResponseRPC, JobListResponseRPC
from renku.ui.service.views import result_response
from renku.ui.service.views.api_versions import ALL_VERSIONS, VersionedBlueprint
from renku.ui.service.views.decorators import requires_cache, requires_identity
from renku.ui.service.views.error_handlers import handle_common_except

JOBS_BLUEPRINT_TAG = "jobs"
jobs_blueprint = VersionedBlueprint("jobs", __name__, url_prefix=SERVICE_PREFIX)


@jobs_blueprint.route("/jobs", methods=["GET"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@requires_cache
@requires_identity
def list_jobs(user_data, cache):
    """
    User created jobs view.

    ---
    get:
      description: Return a listing of jobs for the authenticated user.
      responses:
        200:
          description: List of jobs for the authenticated user.
          content:
            application/json:
              schema: JobListResponseRPC
      tags:
        - jobs
    """
    user = cache.ensure_user(user_data)

    jobs = []
    for job in cache.get_jobs(user):
        try:
            if job.project_id:
                job.project = cache.get_project(user, job.project_id)
        except IntermittentProjectIdError:
            continue

        jobs.append(job)

    return result_response(JobListResponseRPC(), {"jobs": jobs})


@jobs_blueprint.route("/jobs/<job_id>", methods=["GET"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@requires_cache
@requires_identity
def job_details(user_data, cache, job_id):
    """
    Show the details of a specific job.

    ---
    get:
      description: Show the details of a specific job.
      parameters:
        - in: path
          name: job_id
          schema:
            type: string
      responses:
        200:
          description: Details of the job.
          content:
            application/json:
              schema: JobDetailsResponseRPC
      tags:
        - jobs
    """
    user = cache.ensure_user(user_data)
    job = cache.get_job(user, job_id)

    if not job or not job.project_id:
        return result_response(JobDetailsResponseRPC(), None)

    try:
        job.project = cache.get_project(user, job.project_id)
    except IntermittentProjectIdError:
        pass

    return result_response(JobDetailsResponseRPC(), job)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service jobs management."""
from renku.ui.service.cache.base import BaseCache
from renku.ui.service.cache.models.job import Job
from renku.ui.service.cache.models.project import Project
from renku.ui.service.cache.serializers.job import JobSchema


class JobManagementCache(BaseCache):
    """Job management cache."""

    job_schema = JobSchema()

    def make_job(self, user, project=None, job_data=None):
        """Cache job state under user hash set."""
        if job_data:
            job_data.update({"user_id": user.user_id})
        else:
            job_data = {"user_id": user.user_id}

        job_obj = self.job_schema.load(job_data)

        if project and isinstance(project, Project):
            job_obj.project_id = project.project_id
            job_obj.locked.add(project.project_id)

        if project and isinstance(project, str):
            job_obj.project_id = project
            job_obj.locked.add(project)

        job_obj.save()
        return job_obj

    @staticmethod
    def get_job(user, job_id):
        """Get user job."""
        try:
            job_obj = Job.get((Job.job_id == job_id) & (Job.user_id == user.user_id))
        except ValueError:
            return

        return job_obj

    @staticmethod
    def get_jobs(user):
        """Get all user jobs."""
        return Job.query(Job.user_id == user.user_id)

    @staticmethod
    def invalidate_job(user, job_id):
        """Remove users job record."""
        job_obj = JobManagementCache.get_job(user, job_id)

        if job_obj:
            job_obj.delete()

        return job_obj



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service workflow plans view."""
from flask import request

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.workflow_plans_export import WorkflowPlansExportCtrl
from renku.ui.service.controllers.workflow_plans_list import WorkflowPlansListCtrl
from renku.ui.service.controllers.workflow_plans_show import WorkflowPlansShowCtrl
from renku.ui.service.views.api_versions import VERSIONS_FROM_V1_4, VERSIONS_FROM_V1_5, VersionedBlueprint
from renku.ui.service.views.decorators import optional_identity, requires_cache
from renku.ui.service.views.error_handlers import handle_common_except, handle_workflow_errors

PLAN_BLUEPRINT_TAG = "workflow plans"
workflow_plans_blueprint = VersionedBlueprint(PLAN_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX)


@workflow_plans_blueprint.route(
    "/workflow_plans.list", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_FROM_V1_4
)
@handle_common_except
@handle_workflow_errors
@requires_cache
@optional_identity
def list_plans_view(user_data, cache):
    """
    List all plans in a project.

    ---
    get:
      description: List all plans in a project.
      parameters:
        - in: query
          schema: WorkflowPlansListRequest
      responses:
        200:
          description: Listing of all plans in a project.
          content:
            application/json:
              schema: WorkflowPlansListResponseRPC
      tags:
        - workflow plans
    """
    return WorkflowPlansListCtrl(cache, user_data, dict(request.args)).to_response()


@workflow_plans_blueprint.route(
    "/workflow_plans.show", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_FROM_V1_4
)
@handle_common_except
@handle_workflow_errors
@requires_cache
@optional_identity
def show_plan_view(user_data, cache):
    """
    Show details of a plan.

    ---
    get:
      description: Show details of a plan.
      parameters:
        - in: query
          schema: WorkflowPlansShowRequest
      responses:
        200:
          description: The details of the plan.
          content:
            application/json:
              schema:
                WorkflowPlansShowResponseRPC
      tags:
        - workflow plans
    """
    return WorkflowPlansShowCtrl(cache, user_data, dict(request.args)).to_response()


@workflow_plans_blueprint.route(
    "/workflow_plans.export", methods=["POST"], provide_automatic_options=False, versions=VERSIONS_FROM_V1_5
)
@handle_common_except
@handle_workflow_errors
@requires_cache
@optional_identity
def export_plan_view(user_data, cache):
    """
    Export a workflow.

    ---
    post:
      description: Export a workflow to a specific format.
      requestBody:
        content:
          application/json:
            schema: WorkflowPlansExportRequest
      responses:
        200:
          description: The exported plan.
          content:
            application/json:
              schema: WorkflowPlansExportResponseRPC
      tags:
        - workflow plans
    """
    return WorkflowPlansExportCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Curses utilities."""
import curses
import curses.panel
from typing import Dict, List, Optional, Tuple

from renku.command.view_model.text_canvas import Point
from renku.core import errors
from renku.core.util import communication
from renku.domain_model.provenance.activity import Activity


class CursesActivityGraphViewer:
    """An interactive viewer for activity graphs."""

    COLOR_MAPPING = {
        "[31": curses.COLOR_RED,
        "[32": curses.COLOR_GREEN,
        "[33": curses.COLOR_YELLOW,
        "[34": curses.COLOR_BLUE,
        "[35": curses.COLOR_MAGENTA,
        "[36": curses.COLOR_CYAN,
        "[37": curses.COLOR_WHITE,
        "[0": -1,
    }

    ACTIVITY_OVERLAY_WIDTH = 60
    ACTIVITY_OVERLAY_HEIGHT = 35
    HELP_OVERLAY_WIDTH = 60
    HELP_OVERLAY_HEIGHT = 6
    DATE_FORMAT = "%Y-%m-%d %H:%M:S"

    def __init__(
        self,
        text_data: str,
        navigation_data: List[List[Tuple[Point, Point, Activity]]],
        vertical_space: int,
        use_color: bool,
    ):
        self.text_data = text_data
        self.navigation_data = navigation_data
        self.vertical_space = vertical_space

        self.current_layer = 0
        self.layer_position = 0
        self.max_layer = len(navigation_data) - 1
        self.y_pos: int = 0
        self.x_pos: int = 0
        self.color_cache: Dict[str, int] = {}
        self.activity_overlay: Optional[curses._CursesWindow] = None
        self.help_overlay: Optional[curses._CursesWindow] = None
        self._select_activity()
        self.use_color = use_color
        self.free_move: bool = False

    def _init_curses(self, screen):
        """Initialize curses screen for interactive mode."""
        screen.keypad(True)
        curses.use_default_colors()
        curses.start_color()
        curses.curs_set(0)
        curses.noecho()
        screen.refresh()
        self._init_console_colors()
        self.rows, self.cols = screen.getmaxyx()

        min_width = max(self.ACTIVITY_OVERLAY_WIDTH, self.HELP_OVERLAY_WIDTH) + 1
        min_height = max(self.ACTIVITY_OVERLAY_HEIGHT, self.HELP_OVERLAY_HEIGHT) + 1
        if self.cols < min_width or self.rows < min_height:
            raise errors.TerminalSizeError(
                f"Terminal is too small for interactive mode, size should be at least {min_width}x{min_height}."
            )

        text_data_lines = self.text_data.splitlines()

        self.content_max_x = max(len(line) for line in text_data_lines)
        self.content_max_y = len(text_data_lines)

        int16_max = 32767
        if self.content_max_y > int16_max or self.content_max_x > int16_max:
            communication.warn(
                f"Graph is too large for interactive visualization, cropping to {int16_max} lines/columns."
            )
            self.content_max_x = min(self.content_max_x, int16_max)
            self.content_max_y = min(self.content_max_y, int16_max)
            text_data_lines = [line[: self.content_max_x] for line in text_data_lines[self.content_max_y]]

        self.content_pad = curses.newpad(self.content_max_y, self.content_max_x)
        for i, l in enumerate(text_data_lines):
            self._addstr_with_color_codes(self.content_pad, i, 0, l)

        self._blink_text(self.content_pad, self.activity_start, self.activity_end, bold=True)

        self._update_help_overlay(screen)

        self._refresh(screen)

    def _init_console_colors(self):
        """Setup curses color mapping."""
        if not self.use_color:
            curses.init_pair(100, -1, -1)
            for color_symbol, _ in self.COLOR_MAPPING.items():
                self.color_cache[color_symbol] = 100
        else:
            for i, (color_symbol, color) in enumerate(self.COLOR_MAPPING.items(), start=100):
                curses.init_pair(i, color, -1)
                self.color_cache[color_symbol] = i

    def _select_activity(self):
        """Set the currently selected activity."""
        start, end, self.selected_activity = self.navigation_data[self.current_layer][self.layer_position]

        # Ignore borders, we only care about text
        self.activity_start = Point(start.x + 1, start.y + 1)
        self.activity_end = Point(end.x - 1, end.y - 1)

    def _addstr_with_color_codes(self, window, y: int, x: int, text: str):
        """Replace ANSI color codes with curses colors."""
        if not curses.has_colors():
            window.addstr(y, x, text)
            return

        split_text = text.split("\033")

        # add first part without color
        window.addstr(y, x, split_text[0], curses.color_pair(self.color_cache["[0"]))

        x += len(split_text[0])

        for substring in split_text[1:]:
            color, snippet = substring.split("m", 1)

            if len(snippet) == 0:
                continue

            if color == "[1":
                curses_color = curses.A_BOLD
            else:
                curses_color = curses.color_pair(self.color_cache[color])

            window.addstr(y, x, snippet, curses_color)

            x += len(snippet)

    def _blink_text(self, window, start: Point, end: Point, bold: bool = True):
        """Change text between start and end to blinking."""
        style = curses.A_BLINK

        if bold:
            style |= curses.A_BOLD

        for y in range(start.y, end.y + 1):
            text = window.instr(y, start.x, end.x - start.x + 1)
            window.addstr(y, start.x, text, style)

    def _unblink_text(self, window, start: Point, end: Point, bold: bool = True):
        """Change text between start and end to not-blinking."""

        for y in range(start.y, end.y + 1):
            text = window.instr(y, start.x, end.x - start.x + 1)
            if bold:
                window.addstr(y, start.x, text, curses.A_BOLD)
            else:
                window.addstr(y, start.x, text)

    def _add_multiline_text_with_wrapping(self, window, text: str):
        """Add a multiline text to a window, wrapping text to fit."""
        height, width = window.getmaxyx()

        # don't write to window borders
        width -= 2
        i = 1

        for line in text.splitlines():
            chunks = [line[p : p + width] for p in range(0, len(line), width)]
            if not chunks:
                # Lines containing only \n
                i += 1
                continue

            for chunk in chunks:
                if i >= height:
                    # TODO: Add scrolling using a pad?
                    return
                window.addstr(i, 1, chunk)
                i += 1

    def _refresh(self, screen):
        """Refresh curses screens/pads/windows/panels."""
        self.content_pad.refresh(self.y_pos, self.x_pos, 0, 0, self.rows - 1, self.cols - 1)

        if self.activity_overlay:
            self.activity_overlay.overlay(screen)
            self.activity_overlay.refresh()

        if self.help_overlay:
            self.help_overlay.overlay(screen)
            self.help_overlay.refresh()

        if self.activity_overlay or self.help_overlay:
            curses.panel.update_panels()

    def _change_layer(self, step: int):
        """Change the currently active layer."""
        self.current_layer = max(min(self.current_layer + step, self.max_layer), 0)
        self.layer_position = max(min(self.layer_position, len(self.navigation_data[self.current_layer]) - 1), 0)
        del self.activity_overlay
        self.activity_overlay = None

    def _change_layer_position(self, step: int):
        """Change position inside current layer."""
        self.layer_position = max(min(self.layer_position + step, len(self.navigation_data[self.current_layer]) - 1), 0)
        del self.activity_overlay
        self.activity_overlay = None

    def _update_activity_overlay(self, screen):
        """Show/Hide the activity overlay."""
        if not self.activity_overlay:
            del self.help_overlay
            self.help_overlay = None

            self.activity_overlay = curses.newwin(
                self.ACTIVITY_OVERLAY_HEIGHT,
                self.ACTIVITY_OVERLAY_WIDTH,
                self.rows - self.ACTIVITY_OVERLAY_HEIGHT,
                self.cols - self.ACTIVITY_OVERLAY_WIDTH,
            )
            curses.panel.new_panel(self.activity_overlay)
            self.activity_overlay.border()

            started_date = self.selected_activity.started_at_time.strftime(self.DATE_FORMAT)
            ended_date = self.selected_activity.ended_at_time.strftime(self.DATE_FORMAT)
            usages = "\n".join(u.entity.path for u in self.selected_activity.usages)
            generations = "\n".join(g.entity.path for g in self.selected_activity.generations)
            agents = ", ".join(getattr(a, "full_name", a.name) for a in self.selected_activity.agents)
            full_command = " ".join(self.selected_activity.plan_with_values.to_argv(with_streams=True))

            content = (
                "Id:\n"
                f"{self.selected_activity.id}\n\n"
                "Started:\n"
                f"{started_date}\n\n"
                "Ended:\n"
                f"{ended_date}\n\n"
                "Agents:\n"
                f"{agents}\n\n"
                "Plan Id:\n"
                f"{self.selected_activity.association.plan.id}\n\n"
                "Plan Name:\n"
                f"{self.selected_activity.association.plan.name}\n\n"
                "Inputs:\n"
                f"{usages}\n\n"
                "Outputs:\n"
                f"{generations}\n\n"
                "Full Command:\n"
                f"{full_command}\n\n"
            )

            self._add_multiline_text_with_wrapping(self.activity_overlay, content)
            self.activity_overlay.overlay(screen)
        else:
            del self.activity_overlay
            self.activity_overlay = None

    def _update_help_overlay(self, screen):
        """Show/hide the help overlay."""
        if not self.help_overlay:
            del self.activity_overlay
            self.activity_overlay = None

            self.help_overlay = curses.newwin(
                self.HELP_OVERLAY_HEIGHT, self.HELP_OVERLAY_WIDTH, 0, self.cols - self.HELP_OVERLAY_WIDTH
            )
            curses.panel.new_panel(self.help_overlay)
            self.help_overlay.border()

            content = (
                "Navigate using arrow keys\n"
                "Press <enter> to show activity details\n"
                "Press <f> to toggle free arrow movement\n"
                "Press <h> to show/hide this help\n"
                "Press <q> to exit\n"
            )
            self._add_multiline_text_with_wrapping(self.help_overlay, content)
            self.help_overlay.overlay(screen)
        else:
            del self.help_overlay
            self.help_overlay = None

    def _move_viewscreen_to_activity(self):
        """Move viewscreen to include selected activity."""
        if self.activity_start.x - 1 < self.x_pos:
            self.x_pos = max(self.activity_start.x - 1, 0)
        elif self.activity_end.x + 1 > self.x_pos + self.cols:
            self.x_pos = min(self.activity_end.x + 1, self.content_max_x) - self.cols - 1

        if self.activity_start.y - 1 < self.y_pos:
            self.y_pos = max(self.activity_start.y - self.vertical_space - 3, 0)
        elif self.activity_end.y + 1 > self.y_pos + self.rows:
            self.y_pos = min(self.activity_end.y + self.vertical_space + 3, self.content_max_y) - self.rows - 1

    def _loop(self, screen):
        """The interaction loop."""
        running = True

        while running:
            input_char = screen.getch()

            # update screen size to deal with resizes
            self.rows, self.cols = screen.getmaxyx()

            # handle keypress
            if input_char == curses.KEY_DOWN or chr(input_char) == "k":
                if self.free_move:
                    self.y_pos = min(self.y_pos + 1, self.content_max_y - self.rows - 1)
                else:
                    self._change_layer(1)
            elif input_char == curses.KEY_UP or chr(input_char) == "i":
                if self.free_move:
                    self.y_pos = max(self.y_pos - 1, 0)
                else:
                    self._change_layer(-1)
            elif input_char == curses.KEY_RIGHT or chr(input_char) == "l":
                if self.free_move:
                    self.x_pos = min(self.x_pos + 1, self.content_max_x - self.cols - 1)
                else:
                    self._change_layer_position(1)
            elif input_char == curses.KEY_LEFT or chr(input_char) == "j":
                if self.free_move:
                    self.x_pos = max(self.x_pos - 1, 0)
                else:
                    self._change_layer_position(-1)
            elif input_char == curses.KEY_ENTER or input_char == 10 or input_char == 13:
                self._update_activity_overlay(screen)
            elif chr(input_char) == "h":
                self._update_help_overlay(screen)
            elif chr(input_char) == "f":
                self.free_move = not self.free_move
            elif input_char < 256 and chr(input_char) == "q":
                running = False

            self._unblink_text(self.content_pad, self.activity_start, self.activity_end, bold=True)
            self._select_activity()
            self._blink_text(self.content_pad, self.activity_start, self.activity_end, bold=True)

            if not self.free_move:
                self._move_viewscreen_to_activity()

            self._refresh(screen)

    def _main(self, screen):
        """The main execution method for wrapped curses execution."""
        self._init_curses(screen)
        self._loop(screen)

    def run(self):
        """Run interactive curses mode."""
        curses.wrapper(self._main)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project cache management."""
from typing import cast

from marshmallow import EXCLUDE

from renku.ui.service.cache.base import BaseCache
from renku.ui.service.cache.models.project import Project
from renku.ui.service.cache.models.user import User
from renku.ui.service.cache.serializers.project import ProjectSchema
from renku.ui.service.errors import IntermittentProjectIdError


class ProjectManagementCache(BaseCache):
    """Project management cache."""

    project_schema = ProjectSchema()

    def make_project(self, user, project_data, persist=True) -> Project:
        """Store user project metadata."""
        project_data.update({"user_id": user.user_id})

        project_obj: Project = cast(Project, self.project_schema.load(project_data, unknown=EXCLUDE))

        if persist:
            project_obj.save()

        return project_obj

    @staticmethod
    def get_project(user, project_id):
        """Get user cached project."""
        try:
            record = Project.get((Project.project_id == project_id) & (Project.user_id == user.user_id))
        except ValueError as e:
            raise IntermittentProjectIdError(e)

        if not record.abs_path.exists():
            record.delete()
            raise IntermittentProjectIdError()

        return record

    @staticmethod
    def get_projects(user):
        """Get all user cache projects."""
        return Project.query(Project.user_id == user.user_id)

    @staticmethod
    def invalidate_project(user, project_id):
        """Remove user project record."""
        project_obj = ProjectManagementCache.get_project(user, project_id)

        if project_obj:
            project_obj.purge()

        return project_obj

    def user_projects(self):
        """Iterate through all cached projects."""
        for user in User.all():
            yield user, self.get_projects(user)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Remove a file, a directory, or a symlink."""

import os
from pathlib import Path
from subprocess import run
from typing import List, Protocol, runtime_checkable

from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.command.command_builder.command import Command
from renku.core import errors
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.storage import check_external_storage, untrack_paths_from_storage
from renku.core.util import communication
from renku.core.util.git import get_git_user
from renku.core.util.os import delete_dataset_file, expand_directories
from renku.domain_model.project_context import project_context


@runtime_checkable
class EditCommandCallable(Protocol):
    """Typing Protocol for edit command."""

    def __call__(self, filename: str) -> None:
        """The call method."""
        ...


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _remove(sources: List[str], edit_command: EditCommandCallable, dataset_gateway: IDatasetGateway):
    """Remove files and check repository for potential problems.

    Args:
        sources(List[str]): Files to remove.
        edit_command(Callable[[str], None]): Command to execute for editing .gitattributes.
        dataset_gateway(IDatasetGateway): Injected dataset gateway.
    """
    repository = project_context.repository

    def get_relative_path(path):
        """Format path as relative to the project path."""
        abs_path = os.path.abspath(project_context.path / path)
        try:
            return str(Path(abs_path).relative_to(project_context.path))
        except ValueError:
            raise errors.ParameterError(f"File {abs_path} is not within the project.")

    files = {
        get_relative_path(source): get_relative_path(file_or_dir)
        for file_or_dir in sources
        for source in expand_directories((file_or_dir,))
    }

    # 1. Update dataset metadata files.
    progress_text = "Updating dataset metadata"
    all_datasets = dataset_gateway.get_all_active_datasets()
    communication.start_progress(progress_text, total=len(all_datasets))
    try:
        for dataset in all_datasets:
            remove = []
            for file in dataset.files:
                key = file.entity.path
                filepath = get_relative_path(key)
                if filepath in files:
                    remove.append(key)

            if remove:
                dataset = dataset.copy()
                for key in remove:
                    dataset.unlink_file(key)
                    delete_dataset_file(project_context.path / key, follow_symlinks=True)

                datasets_provenance = DatasetsProvenance()
                datasets_provenance.add_or_update(dataset, creator=get_git_user(repository))
            communication.update_progress(progress_text, amount=1)
    finally:
        communication.finalize_progress(progress_text)

    # 2. Manage .gitattributes for external storage.
    if check_external_storage():
        tracked = tuple(path for path, attr in repository.get_attributes(*files).items() if attr.get("filter") == "lfs")
        untrack_paths_from_storage(*tracked)
        existing = repository.get_attributes(*tracked)
        if existing:
            communication.warn("There are custom .gitattributes.\n")
            if communication.confirm('Do you want to edit ".gitattributes" now?'):
                edit_command(filename=str(project_context.path / ".gitattributes"))

    # Finally remove the files.
    files_to_remove = {str(project_context.path / f) for f in files.values()}
    final_sources = list(files_to_remove)
    if final_sources:
        run(["git", "rm", "-rf"] + final_sources, check=True)


def remove_command():
    """Command to remove a file."""
    return Command().command(_remove).require_clean().with_database(write=True).with_commit()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Template models."""

import copy
import json
import os
import tempfile
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Set, Tuple, Union, cast

import jinja2
import yaml

from renku.core import errors
from renku.core.constant import RENKU_HOME
from renku.core.util.os import get_safe_relative_path, hash_file, hash_string
from renku.core.util.util import to_string

if TYPE_CHECKING:
    from renku.domain_model.project import Project

TEMPLATE_MANIFEST = "manifest.yaml"


class TemplatesSource:
    """Base class for Renku template sources."""

    def __init__(self, path, source, reference, version, skip_validation: bool = False):
        self.path: Path = Path(path)
        self.source: str = source
        self.reference: Optional[str] = reference
        self.version: str = version
        self.manifest: TemplatesManifest = TemplatesManifest.from_path(path / TEMPLATE_MANIFEST, skip_validation)

    @classmethod
    @abstractmethod
    def fetch(cls, source: Optional[str], reference: Optional[str]) -> "TemplatesSource":
        """Fetch an embedded or remote template."""
        raise NotImplementedError

    @property
    def templates(self) -> List["Template"]:
        """Return list of templates."""
        for template in self.manifest.templates:
            template.templates_source = self
            template.validate(skip_files=False)

        return self.manifest.templates

    def is_update_available(
        self, id: str, reference: Optional[str], version: Optional[str]
    ) -> Tuple[bool, Optional[str]]:
        """Return True if an update is available along with the latest reference of a template."""
        latest = self.get_latest_reference_and_version(id=id, reference=reference, version=version)
        if not latest:
            return False, reference

        latest_reference, latest_version = latest
        update_available = latest_reference != reference or latest_version != version
        return update_available, latest_reference

    @abstractmethod
    def get_all_references(self, id) -> List[str]:
        """Return all available versions for a template id."""
        raise NotImplementedError

    @abstractmethod
    def get_latest_reference_and_version(
        self, id: str, reference: Optional[str], version: Optional[str]
    ) -> Optional[Tuple[Optional[str], str]]:
        """Return latest reference and version number of a template."""
        raise NotImplementedError

    @abstractmethod
    def get_template(self, id, reference: Optional[str]) -> "Template":
        """Return a template at a specific reference."""
        raise NotImplementedError


class TemplatesManifest:
    """Manifest file for Renku templates."""

    def __init__(self, content: List[Dict], skip_validation: bool = False):
        self._content: List[Dict] = content
        self._templates: Optional[List[Template]] = None

        if not skip_validation:
            self.validate()

    @classmethod
    def from_path(cls, path: Union[Path, str], skip_validation: bool = False) -> "TemplatesManifest":
        """Extract template metadata from the manifest file."""
        try:
            return cls.from_string(Path(path).read_text(), skip_validation)
        except FileNotFoundError as e:
            raise errors.InvalidTemplateError(f"There is no manifest file '{path}'") from e
        except UnicodeDecodeError as e:
            raise errors.InvalidTemplateError(f"Cannot read manifest file '{path}'") from e

    @classmethod
    def from_string(cls, content: str, skip_validation: bool = False) -> "TemplatesManifest":
        """Extract template metadata from the manifest file."""
        try:
            manifest = yaml.safe_load(content)
        except yaml.YAMLError as e:
            raise errors.InvalidTemplateError("Cannot parse manifest file") from e
        else:
            manifest = TemplatesManifest(manifest, skip_validation)
            return manifest

    @property
    def templates(self) -> List["Template"]:
        """Return list of available templates info in the manifest."""
        if self._templates is None:
            self._templates = [
                Template(
                    id=cast(str, t.get("id") or t.get("folder")),
                    aliases=t.get("aliases", []),
                    name=cast(str, t.get("name")),
                    description=cast(str, t.get("description")),
                    parameters=cast(Dict[str, Dict[str, Any]], t.get("variables") or t.get("parameters")),
                    icon=cast(str, t.get("icon")),
                    ssh_supported=t.get("ssh_supported", False),
                    immutable_files=t.get("immutable_template_files", []),
                    allow_update=t.get("allow_template_update", True),
                    source=None,
                    reference=None,
                    version=None,
                    path=None,
                    templates_source=None,
                )
                for t in self._content
            ]

        return self._templates

    def get_raw_content(self) -> List[Dict]:
        """Return raw manifest file content."""
        return copy.deepcopy(self._content)

    def validate(self, manifest_only: bool = False) -> List[str]:
        """Validate manifest content."""
        warnings = []

        if not self._content:
            raise errors.InvalidTemplateError("Cannot find any valid template in manifest file")
        elif not isinstance(self._content, list):
            raise errors.InvalidTemplateError(f"Invalid manifest content type: '{type(self._content).__name__}'")

        existing_ids: Set[str] = set()

        # NOTE: First check if required fields exists for creating Template instances
        for template_entry in self._content:
            if not isinstance(template_entry, dict):
                raise errors.InvalidTemplateError(f"Invalid template type: '{type(template_entry).__name__}'")

            id = template_entry.get("id") or template_entry.get("folder")
            if not id:
                raise errors.InvalidTemplateError(f"Template doesn't have an id: '{template_entry}'")
            if not template_entry.get("id"):
                warnings.append(f"Template '{id}' should use 'id' attribute instead of 'folder'.")

            # NOTE: Check for duplicate IDs and aliases
            aliases = {id}
            aliases.update(template_entry.get("aliases", []))
            duplicates = existing_ids.intersection(aliases)
            if duplicates:
                duplicates_str = ", ".join(sorted(f"'{d}'" for d in duplicates))
                raise errors.InvalidTemplateError(f"Found duplicate IDs or aliases: {duplicates_str}")
            existing_ids.update(aliases)

            parameters = template_entry.get("variables")
            if parameters:
                if not isinstance(parameters, dict):
                    raise errors.InvalidTemplateError(
                        f"Invalid template variable type on template '{id}': '{type(parameters).__name__}', "
                        "should be 'dict'."
                    )

                for key, parameter in parameters.items():
                    if isinstance(parameter, str):  # NOTE: Backwards compatibility
                        template_entry["variables"][key] = {"description": parameter}
                        warnings.append(
                            f"Template '{id}' variable '{key}' uses old string format in manifest and should be"
                            " replaced with the nested dictionary format."
                        )
        if not manifest_only:
            for template in self.templates:
                template.validate(skip_files=True)

        return warnings


class Template:
    """Template files and metadata from a template source."""

    REQUIRED_ATTRIBUTES = ("name",)
    REQUIRED_FILES = (os.path.join(RENKU_HOME, "renku.ini"), "Dockerfile")
    PROHIBITED_PATHS = (f"{RENKU_HOME}/*",)

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        parameters: Dict[str, Dict[str, Any]],
        icon: str,
        ssh_supported: bool,
        aliases: List[str],
        immutable_files: List[str],
        allow_update: bool,
        source: Optional[str],
        reference: Optional[str],
        version: Optional[str],
        path: Optional[Path],
        templates_source: Optional[TemplatesSource],
    ):
        self.path: Optional[Path] = path
        self.source = source
        self.reference = reference
        self.version = version
        self.id: str = id
        self.name: str = name
        self.description: str = description
        self.icon = icon
        self.ssh_supported = ssh_supported
        self.aliases: List[str] = aliases
        self.immutable_files: List[str] = immutable_files or []
        self.allow_update: bool = allow_update
        parameters = parameters or {}
        self.parameters: List[TemplateParameter] = [
            TemplateParameter.from_dict(name=k, value=v) for k, v in parameters.items()
        ]

        self._templates_source: Optional[TemplatesSource] = templates_source

    def __repr__(self) -> str:
        return f"<Template {self.id}@{self.version}>"

    @property
    def templates_source(self) -> Optional[TemplatesSource]:
        """Return template's source."""
        return self._templates_source

    @templates_source.setter
    def templates_source(self, templates_source: TemplatesSource):
        """Set templates source for this template."""
        self._templates_source = templates_source
        self.source = templates_source.source
        self.reference = templates_source.reference
        self.version = templates_source.version
        self.path = templates_source.path / self.id

    def get_all_references(self) -> List[str]:
        """Return all available references for the template."""
        if self.templates_source is None:
            return []
        return self.templates_source.get_all_references(self.id)

    def validate(self, skip_files: bool, raise_errors: bool = True) -> List[str]:
        """Validate a template."""

        issues = []

        for attribute in self.REQUIRED_ATTRIBUTES:
            if not getattr(self, attribute, None):
                issue = f"Template '{self.id}' does not have a '{attribute}' attribute"
                if raise_errors:
                    raise errors.InvalidTemplateError(issue)
                issues.append(issue)

        for parameter in self.parameters:
            issues.extend(parameter.validate(raise_errors=False))

        if skip_files:
            return issues

        if self.path is None or not self.path.exists():
            issue = f"Template directory for '{self.id}' does not exists"
            if raise_errors:
                raise errors.InvalidTemplateError(issue)
            issues.append(issue)
            return issues  # NOTE: no point checking individual files if directory doesn't exist.

        missing_required_files: Set[str] = set()
        for file in self.REQUIRED_FILES:
            if not (self.path / file).is_file():
                missing_required_files.add(file)

        if missing_required_files:
            required_files_str = "\n\t\t\t".join(missing_required_files)
            issue = f"These paths are required but missing:\n\t\t\t{required_files_str}"
            if raise_errors:
                raise errors.InvalidTemplateError(issue)
            issues.append(issue)

        existing_prohibited_paths: Set[str] = set()

        for pattern in self.PROHIBITED_PATHS:
            matches = {m for m in self.path.glob(pattern) if str(m.relative_to(self.path)) not in self.REQUIRED_FILES}
            if matches:
                existing_prohibited_paths.update(str(m.relative_to(self.path)) for m in matches)

        if existing_prohibited_paths:
            prohibited_paths_str = "\n\t\t\t".join(p for p in existing_prohibited_paths)
            issue = f"These paths are not allowed in a template:\n\t\t\t{prohibited_paths_str}"
            if raise_errors:
                raise errors.InvalidTemplateError(issue)
            issues.append(issue)

        # NOTE: Validate symlinks resolve to a path inside the template
        for relative_path in self.get_files():
            try:
                get_safe_relative_path(path=relative_path, base=self.path)
            except ValueError:
                issue = f"File '{relative_path}' is not within the template."
                if raise_errors:
                    raise errors.InvalidTemplateError(issue)
                issues.append(issue)

        return issues

    def get_files(self) -> Generator[str, None, None]:
        """Return all files in a rendered renku template."""
        if self.path is None:
            return
        for subpath in self.path.rglob("*"):
            if subpath.is_file():
                yield str(subpath.relative_to(self.path))

    def render(self, metadata: "TemplateMetadata") -> "RenderedTemplate":
        """Render template files in a new directory."""
        if self.path is None:
            raise ValueError("Template path not set")

        render_base = Path(tempfile.mkdtemp())

        for relative_path in self.get_files():
            # NOTE: The path could contain template variables, we need to template it
            rendered_relative_path = jinja2.Template(relative_path).render(metadata.metadata)

            destination = render_base / rendered_relative_path
            destination.parent.mkdir(parents=True, exist_ok=True)

            source = self.path / relative_path

            try:
                content = source.read_text()
            except UnicodeDecodeError:  # NOTE: Binary files
                content_bytes = source.read_bytes()
                destination.write_bytes(content_bytes)
            else:
                template = jinja2.Template(content, keep_trailing_newline=True)
                rendered_content = template.render(metadata.metadata)
                destination.write_text(rendered_content)

        return RenderedTemplate(path=render_base, template=self, metadata=metadata.metadata)


class RenderedTemplate:
    """A rendered version of a Template."""

    def __init__(self, path: Path, template: Template, metadata: Dict[str, Any]):
        self.path: Path = path
        self.template: Template = template
        self.metadata: Dict[str, Any] = metadata
        self.checksums: Dict[str, Optional[str]] = {
            f: hash_template_file(relative_path=f, absolute_path=self.path / f) for f in self.get_files()
        }

    def get_files(self) -> Generator[str, None, None]:
        """Return all files in a rendered renku template."""
        for subpath in self.path.rglob("*"):
            if not subpath.is_file():
                continue

            relative_path = str(subpath.relative_to(self.path))

            yield relative_path


class TemplateParameter:
    """Represent template variables."""

    VALID_TYPES = ("string", "number", "boolean", "enum")

    def __init__(
        self,
        name: str,
        description: Optional[str],
        type: Optional[str],
        possible_values: Optional[List[Union[int, float, str, bool]]],
        default: Optional[Union[int, float, str, bool]],
    ):
        self.name: str = name
        self.description: str = description or ""
        self.type: Optional[str] = type
        self.possible_values: List[Union[int, float, str, bool]] = possible_values or []
        self.default = default

    @classmethod
    def from_dict(cls, name: str, value: Dict[str, Any]):
        """Create an instance from a dict."""
        if not name:
            raise errors.InvalidTemplateError(f"No name specified for template parameter '{value}'")
        if not isinstance(value, dict):
            raise errors.InvalidTemplateError(f"Invalid parameter type '{type(value).__name__}' for '{name}'")

        return cls(
            name=name,
            type=value.get("type"),
            description=value.get("description"),
            possible_values=value.get("possible_values") or value.get("enum"),
            default=value.get("default_value"),
        )

    @property
    def has_default(self) -> bool:
        """Return True if a default value is set."""
        # NOTE: ``None`` cannot be used as the default value but it's ok since no variable type accepts it and it's not
        # a valid value anyways
        return self.default is not None

    def validate(self, raise_errors: bool = True) -> List[str]:
        """Validate manifest content."""
        issues = []
        if not self.name:
            issue = "Template parameter does not have a name."
            if raise_errors:
                raise errors.InvalidTemplateError(issue)
            issues.append(issue)

        if self.type and self.type not in self.VALID_TYPES:
            issue = f"Template contains variable '{self.name}' of type '{self.type}' which is not supported"
            if raise_errors:
                raise errors.InvalidTemplateError(issue)
            issues.append(issue)

        if self.possible_values and not isinstance(self.possible_values, list):
            issue = (  # type: ignore[unreachable]
                "Invalid type for possible values of template variable " f"'{self.name}': '{self.possible_values}'"
            )
            if raise_errors:
                raise errors.InvalidTemplateError(issue)
            issues.append(issue)

        if self.type and self.type == "enum" and not self.possible_values:
            issue = f"Template variable '{self.name}' of type enum does not provide a corresponding enum list"
            if raise_errors:
                raise errors.InvalidTemplateError(issue)
            issues.append(issue)

        if self.has_default:
            try:
                self.default = self.convert(self.default)  # type: ignore
            except ValueError as e:
                issue = f"Invalid default value for '{self.name}': {e}"
                if raise_errors:
                    raise errors.InvalidTemplateError(issue)
                issues.append(issue)
        return issues

    def convert(self, value: Union[int, float, str, bool]) -> Union[int, float, str, bool]:
        """Convert a given value to the proper type and raise if value is not valid."""
        valid = True

        if not self.type:
            return value
        elif self.type == "string":
            if not isinstance(value, str):
                valid = False
        elif self.type == "number":
            try:
                value = int(str(value))  # NOTE: Convert to str first to avoid converting float to int if value is float
            except ValueError:
                try:
                    value = float(value)
                except ValueError:
                    valid = False
        elif self.type == "boolean":
            true = (True, 1, "1", "true", "True")
            false = (False, 0, "0", "false", "False")
            if value not in true and value not in false:
                valid = False
            else:
                value = True if value in true else False
        elif self.type == "enum":
            if value not in self.possible_values:
                valid = False

        if not valid:
            value = f"{value} (type: {type(value).__name__})"
            kind = f"type: {self.type}" if self.type else ""
            possible_values = f"possible values: {self.possible_values}" if self.possible_values else ""
            separator = ", " if kind and possible_values else ""
            info = f" ({kind}{separator}{possible_values})" if kind or possible_values else ""
            raise ValueError(f"Invalid value '{value}' for template variable '{self.name}{info}'")

        return value


class TemplateMetadata:
    """Metadata required for rendering a template."""

    def __init__(self, metadata: Dict[str, Any], immutable_files: List[str]):
        self.metadata: Dict[str, Any] = metadata or {}
        self.immutable_files: List[str] = immutable_files or []

    @classmethod
    def from_dict(cls, metadata: Dict[str, Any]) -> "TemplateMetadata":
        """Return an instance from a metadata dict."""
        return cls(metadata=metadata, immutable_files=[])

    @classmethod
    def from_project(cls, project: Optional["Project"]) -> "TemplateMetadata":
        """Return an instance from reading template-related metadata from a project."""
        from renku.core.util.metadata import read_renku_version_from_dockerfile
        from renku.version import __version__

        if not project:
            metadata = {}
            immutable_files: List[str] = []
        else:
            metadata = json.loads(project.template_metadata.metadata) if project.template_metadata.metadata else {}

            # NOTE: Make sure project's template metadata is updated
            metadata["__template_source__"] = project.template_metadata.template_source
            metadata["__template_ref__"] = project.template_metadata.template_ref
            metadata["__template_version__"] = project.template_metadata.template_version
            metadata["__template_id__"] = project.template_metadata.template_id
            # NOTE: Ignore Project.automated_update since it's default is False and won't allow any update at all

            immutable_files = project.template_metadata.immutable_template_files or []

        # NOTE: Always set __renku_version__ to the value read from the Dockerfile (if available) since setting/updating
        # the template doesn't change project's metadata version and shouldn't update the Renku version either
        renku_version = metadata.get("__renku_version__")
        dockerfile_version = to_string(read_renku_version_from_dockerfile())
        metadata["__renku_version__"] = dockerfile_version or renku_version or __version__

        return cls(metadata=metadata, immutable_files=immutable_files)

    @property
    def source(self):
        """Template source."""
        return self.metadata.get("__template_source__")

    @property
    def reference(self):
        """Template reference."""
        return self.metadata.get("__template_ref__")

    @property
    def version(self):
        """Template version."""
        return self.metadata.get("__template_version__")

    @property
    def id(self):
        """Template id."""
        return self.metadata.get("__template_id__")

    @property
    def allow_update(self) -> bool:
        """Is template updatable."""
        return self.metadata.get("__automated_update__", True)

    def update(self, template: Template):
        """Update metadata from a template."""
        self.metadata["__template_source__"] = template.source
        self.metadata["__template_ref__"] = template.reference
        self.metadata["__template_version__"] = template.version
        self.metadata["__template_id__"] = template.id
        self.metadata["__automated_update__"] = template.allow_update
        self.immutable_files = template.immutable_files


def find_renku_section(lines: List[str]) -> Tuple[int, int]:
    """Return start and end line numbers of the Renku-specific section."""
    start = end = -1
    for index, line in enumerate(lines):
        if line.startswith("#        Renku-specific section - DO NOT MODIFY        #"):
            start = index
        elif line.endswith("#              End Renku-specific section              #"):
            end = index
            break

    return start, end


def get_renku_section_from_dockerfile(content: str) -> Optional[str]:
    """Return the Renku-specific section of the Dockerfile or the whole Dockerfile if it doesn't exist."""
    lines = [line.rstrip() for line in content.splitlines()]
    start, end = find_renku_section(lines)

    if 0 <= start < end:
        lines = lines[start:end]
        lines = [line for line in lines if line]  # NOTE: Remove empty lines
        return "\n".join(lines)
    else:
        return None


def calculate_dockerfile_checksum(
    *, dockerfile: Optional[Path] = None, dockerfile_content: Optional[str] = None
) -> str:
    """Calculate checksum for the given file or content.

    NOTE: We ignore empty lines and whitespace characters at the end of the lines when calculating Dockerfile checksum
    if it has Renku-specific section markers.
    """
    if not dockerfile and not dockerfile_content:
        raise errors.ParameterError("Either Dockerfile or its content must be passed")
    elif dockerfile and dockerfile_content:
        raise errors.ParameterError("Cannot pass both Dockerfile and its content")

    content = dockerfile_content if dockerfile_content is not None else dockerfile.read_text()  # type: ignore
    renku_section = get_renku_section_from_dockerfile(content) or content
    return hash_string(renku_section)


def update_dockerfile_content(source: Path, destination: Path) -> None:
    """Update the Renku-specific section of the destination Dockerfile with the one from the source Dockerfile."""
    source_lines = [line.rstrip() for line in source.read_text().splitlines()]
    source_start, source_end = find_renku_section(source_lines)

    destination_lines = [line.rstrip() for line in destination.read_text().splitlines()]
    destination_start, destination_end = find_renku_section(destination_lines)

    # NOTE: If source or destination Dockerfiles doesn't have Renku-specific section, we overwrite the whole file
    if 0 <= source_start < source_end and 0 <= destination_start < destination_end:
        destination_lines[destination_start:destination_end] = source_lines[source_start:source_end]
        content = "\n".join(destination_lines)
        destination.write_text(content)
    else:
        destination.write_text(source.read_text())


def hash_template_file(*, relative_path: Union[Path, str], absolute_path: Union[Path, str]) -> Optional[str]:
    """Use proper hash on a template file."""
    return (
        calculate_dockerfile_checksum(dockerfile=Path(absolute_path))
        if str(relative_path) == "Dockerfile"
        else hash_file(absolute_path)
    )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Workflow file parser."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Tuple, Union

if TYPE_CHECKING:
    from renku.core.workflow.model.workflow_file import WorkflowFile


class IWorkflowFileParser(metaclass=ABCMeta):
    """Abstract class for parsing a workflow file."""

    @abstractmethod
    def workflow_file_parser(self) -> Tuple[IWorkflowFileParser, str]:
        """Supported workflow description formats.

        Returns:
            Tuple[IWorkflowParser, str]: A tuple of the provider itself and the workflow parser name.
        """
        raise NotImplementedError

    @abstractmethod
    def parse(self, path: Union[Path, str]) -> WorkflowFile:
        """Parse a given workflow file using the provider.

        Args:
            path(Union[Path, str]): Path to the workflow file to parse.

        Returns:
            WorkflowFile: A ``WorkflowFile`` generated by the provider.
        """
        raise NotImplementedError



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command builder for local object database."""


import json
import os
from typing import Optional

from packaging.version import Version

from renku.command.command_builder.command import Command, CommandResult, check_finalized
from renku.core import errors
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.database_gateway import IDatabaseGateway
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.interface.storage import IStorageFactory
from renku.domain_model.project_context import project_context
from renku.infrastructure.gateway.activity_gateway import ActivityGateway
from renku.infrastructure.gateway.database_gateway import DatabaseGateway
from renku.infrastructure.gateway.dataset_gateway import DatasetGateway
from renku.infrastructure.gateway.plan_gateway import PlanGateway
from renku.infrastructure.gateway.project_gateway import ProjectGateway
from renku.infrastructure.storage.factory import StorageFactory


class DatabaseCommand(Command):
    """Builder to get a database connection."""

    PRE_ORDER = 4
    POST_ORDER = 5

    def __init__(self, builder: Command, write: bool = False, path: Optional[str] = None, create: bool = False) -> None:
        self._builder = builder
        self._write = write
        self._path = path
        self._create = create
        self.project_found: bool = False

    def _injection_pre_hook(self, builder: Command, context: dict, *args, **kwargs) -> None:
        """Create a Database singleton."""
        from renku.version import __version__

        if not project_context.has_context():
            raise ValueError("Database builder needs a ProjectContext to be set.")

        project_context.push_path(path=self._path or project_context.path, save_changes=self._write)

        project_gateway = ProjectGateway()

        context["constructor_bindings"][IPlanGateway] = lambda: PlanGateway()
        context["constructor_bindings"][IActivityGateway] = lambda: ActivityGateway()
        context["constructor_bindings"][IDatabaseGateway] = lambda: DatabaseGateway()
        context["constructor_bindings"][IDatasetGateway] = lambda: DatasetGateway()
        context["constructor_bindings"][IProjectGateway] = lambda: project_gateway
        context["constructor_bindings"][IStorageFactory] = lambda: StorageFactory

        if int(os.environ.get("RENKU_SKIP_MIN_VERSION_CHECK", "0")) == 1:
            # NOTE: Used for unit tests
            return

        try:
            project = project_gateway.get_project()
            minimum_renku_version = Version(project.minimum_renku_version)
            self.project_found = True
        except (KeyError, ImportError, ValueError):
            try:
                with open(project_context.database_path / "project") as f:
                    project = json.load(f)
                    min_version = project.get("minimum_renku_version")
                    if min_version is None:
                        return
                    minimum_renku_version = Version(min_version)
            except (KeyError, OSError, json.JSONDecodeError):
                # NOTE: We don't check minimum version if there's no project metadata available
                return

        current_version = Version(__version__)

        if Version(current_version.base_version) < minimum_renku_version:
            raise errors.MinimumVersionError(current_version, minimum_renku_version)

    def _post_hook(self, builder: Command, context: dict, result: CommandResult, *args, **kwargs) -> None:
        from renku.domain_model.project import Project

        if self._write and self.project_found:
            # NOTE: Fetch project again in case it was updated (the current reference would be put of date)
            project_gateway = ProjectGateway()
            project = project_gateway.get_project()

            if Version(project.minimum_renku_version) < Version(Project.minimum_renku_version):
                # NOTE: update minimum renku version on write as migrations might happen on the fly
                project.minimum_renku_version = Project.minimum_renku_version

        project_context.pop_context()

    @check_finalized
    def build(self) -> Command:
        """Build the command."""
        self._builder.add_injection_pre_hook(self.PRE_ORDER, self._injection_pre_hook)
        self._builder.add_post_hook(self.POST_ORDER, self._post_hook)

        return self._builder.build()



#  Copyright Swiss Data Science Center (SDSC). A partnership between
#  École Polytechnique Fédérale de Lausanne (EPFL) and
#  Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions used for sessions."""
import os
import urllib
from typing import Optional

from renku.core.util.git import get_remote
from renku.core.util.urls import parse_authentication_endpoint
from renku.domain_model.project_context import project_context


def get_renku_project_name() -> str:
    """Get the full name of a renku project."""
    repository = project_context.repository

    project_name = (
        f"{project_context.remote.owner}/{project_context.remote.name}"
        if project_context.remote.name
        else f"{project_context.path.name}"
    )
    if get_remote(repository, name="renku-backup-origin") and project_name.startswith("repos/"):
        project_name = project_name.replace("repos/", "", 1)
    return project_name


def get_renku_url() -> Optional[str]:
    """Derive the URL of the Renku deployment."""
    renku_url = parse_authentication_endpoint(use_remote=True)
    if renku_url:
        renku_url = urllib.parse.urlunparse(renku_url)
    return renku_url


def get_image_repository_host() -> Optional[str]:
    """Derive the hostname for the gitlab container registry."""
    # NOTE: Used to circumvent cases where the registry URL is not guessed correctly
    # remove once #3301 is done
    if "RENKU_IMAGE_REGISTRY" in os.environ:
        return os.environ["RENKU_IMAGE_REGISTRY"]
    renku_url = get_renku_url()
    if not renku_url:
        return None
    return "registry." + urllib.parse.urlparse(renku_url).netloc



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataverse API integration."""

import json
import posixpath
import re
import urllib
from pathlib import Path
from string import Template
from typing import TYPE_CHECKING, Any, Dict, List, Optional
from urllib import parse as urlparse

from renku.core import errors
from renku.core.config import get_value, set_value
from renku.core.dataset.providers.api import (
    ExporterApi,
    ExportProviderInterface,
    ImportProviderInterface,
    ProviderApi,
    ProviderPriority,
)
from renku.core.dataset.providers.dataverse_metadata_templates import (
    AUTHOR_METADATA_TEMPLATE,
    CONTACT_METADATA_TEMPLATE,
    DATASET_METADATA_TEMPLATE,
    KEYWORDS_METADATA_TEMPLATE,
)
from renku.core.dataset.providers.doi import DOIProvider
from renku.core.dataset.providers.repository import RepositoryImporter, make_request
from renku.core.util import communication
from renku.core.util.datetime8601 import fix_datetime
from renku.core.util.doi import extract_doi, get_doi_url, is_doi
from renku.core.util.urls import remove_credentials
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import ProviderDataset, ProviderParameter
    from renku.domain_model.dataset import Dataset, DatasetTag

DATAVERSE_API_PATH = "api/v1"

DATAVERSE_VERSION_API = "info/version"
DATAVERSE_METADATA_API = "datasets/export"
DATAVERSE_VERSIONS_API = "datasets/:persistentId/versions"
DATAVERSE_FILE_API = "access/datafile/:persistentId/"
DATAVERSE_EXPORTER = "schema.org"

DATAVERSE_SUBJECTS = [
    "Agricultural Sciences",
    "Arts and Humanities",
    "Astronomy and Astrophysics",
    "Business and Management",
    "Chemistry",
    "Computer and Information Science",
    "Earth and Environmental Sciences",
    "Engineering",
    "Law",
    "Mathematical Sciences",
    "Medicine, Health and Life Sciences",
    "Physics",
    "Social Sciences",
    "Other",
]


class DataverseProvider(ProviderApi, ExportProviderInterface, ImportProviderInterface):
    """Dataverse API provider."""

    priority = ProviderPriority.HIGH
    name = "Dataverse"
    is_remote = True

    def __init__(self, uri: str, is_doi: bool = False):
        super().__init__(uri=uri)

        self.is_doi = is_doi
        self._server_url = None
        self._dataverse_name = None
        self._publish: bool = False

    @staticmethod
    def supports(uri):
        """Check if provider supports a given URI."""
        is_doi_ = is_doi(uri)

        is_dataverse_uri = is_doi_ is None and check_dataverse_uri(uri)
        is_dataverse_doi = is_doi_ and check_dataverse_doi(is_doi_.group(0))

        return is_dataverse_uri or is_dataverse_doi

    @staticmethod
    def get_export_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for export."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [
            ProviderParameter("dataverse-server", help="Dataverse server URL.", type=str),
            ProviderParameter("dataverse-name", help="Dataverse name to export to.", type=str),
            ProviderParameter("publish", help="Publish the exported dataset.", is_flag=True),
        ]

    @staticmethod
    def record_id(uri):
        """Extract record id from URI."""
        parsed = urlparse.urlparse(uri)
        return urlparse.parse_qs(parsed.query)["persistentId"][0]

    def get_importer(self, **kwargs) -> "DataverseImporter":
        """Get importer for a record from Dataverse.

        Returns:
            DataverseImporter: The found record
        """

        def get_export_uri(uri):
            """Gets a dataverse api export URI from a dataverse entry."""
            record_id = DataverseProvider.record_id(uri)
            return make_records_url(record_id, uri)

        uri = self.uri
        if self.is_doi:
            doi = DOIProvider(uri=uri).get_importer()
            uri = doi.uri

        uri = get_export_uri(uri)
        response = make_request(uri)

        return DataverseImporter(json=response.json(), uri=uri, original_uri=self.uri)

    def get_exporter(
        self,
        dataset: "Dataset",
        *,
        tag: Optional["DatasetTag"],
        dataverse_server: Optional[str] = None,
        dataverse_name: Optional[str] = None,
        publish: bool = False,
        **kwargs,
    ) -> "ExporterApi":
        """Create export manager for given dataset."""

        def set_export_parameters():
            """Set and validate required parameters for exporting for a provider."""

            server = dataverse_server
            config_base_url = "server_url"
            if not server:
                server = get_value("dataverse", config_base_url)
            else:
                set_value("dataverse", config_base_url, server, global_only=True)

            if not server:
                raise errors.ParameterError("Dataverse server URL is required.")

            if not dataverse_name:
                raise errors.ParameterError("Dataverse name is required.")

            self._server_url = server  # type: ignore
            self._dataverse_name = dataverse_name  # type: ignore
            self._publish = publish

        set_export_parameters()
        return DataverseExporter(dataset=dataset, server_url=self._server_url, dataverse_name=self._dataverse_name)


class DataverseImporter(RepositoryImporter):
    """Dataverse record serializer."""

    def __init__(self, uri: str, original_uri: str, json: Dict[str, Any]):
        super().__init__(uri=uri, original_uri=original_uri)
        self._json: Dict[str, Any] = json

    def is_latest_version(self):
        """Check if record is at last possible version."""
        return True

    @staticmethod
    def _convert_json_property_name(property_name):
        """Removes '@' and converts names to snake_case."""
        property_name = property_name.strip("@")
        property_name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", property_name)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", property_name).lower()

    @property
    def version(self):
        """Get the major and minor version of this dataset."""
        uri = make_versions_url(DataverseProvider.record_id(self._uri), self._uri)
        response = make_request(uri).json()
        newest_version = response["data"][0]
        return "{}.{}".format(newest_version["versionNumber"], newest_version["versionMinorNumber"])

    @property
    def latest_uri(self):
        """Get URI of latest version."""
        return self._uri

    def get_files(self):
        """Get Dataverse files metadata as ``DataverseFileSerializer``."""
        files = []

        for f in self._json["distribution"]:
            mapped_file = {self._convert_json_property_name(k): v for k, v in f.items()}
            mapped_file["parent_url"] = self._uri
            files.append(mapped_file)

        if not files:
            raise LookupError("no files have been found - deposit is empty or protected")

        return [DataverseFileSerializer(**file) for file in files]

    def fetch_provider_dataset(self) -> "ProviderDataset":
        """Deserialize a ``Dataset``."""
        from marshmallow import post_load, pre_load

        from renku.command.schema.agent import PersonSchema
        from renku.core.dataset.providers.models import ProviderDataset, ProviderDatasetFile, ProviderDatasetSchema
        from renku.domain_model.dataset import Url, generate_default_slug

        class DataverseDatasetSchema(ProviderDatasetSchema):
            """Schema for Dataverse datasets."""

            @pre_load
            def fix_data(self, data, **kwargs):
                """Fix data that is received from Dataverse."""
                # Fix context
                context = data.get("@context")
                if context and isinstance(context, str):
                    if context == "http://schema.org":
                        context = "http://schema.org/"
                    data["@context"] = {"@base": context, "@vocab": context}

                # Add type to creators
                creators = data.get("creator", [])
                for c in creators:
                    c["@type"] = [str(t) for t in PersonSchema.opts.rdf_type]

                # Fix license to be a string
                license = data.get("license")
                if license and isinstance(license, dict):
                    data["license"] = license.get("url", "")

                return data

            @post_load
            def fix_timezone(self, obj, **kwargs):
                """Add timezone to datetime object."""
                if obj.get("date_modified"):
                    obj["date_modified"] = fix_datetime(obj["date_modified"])
                if obj.get("date_published"):
                    obj["date_published"] = fix_datetime(obj["date_published"])

                return obj

        files = self.get_files()
        dataset = ProviderDataset.from_jsonld(data=self._json, schema_class=DataverseDatasetSchema)
        dataset.version = self.version
        dataset.slug = generate_default_slug(name=dataset.name or "", version=dataset.version)
        dataset.same_as = (
            Url(url_str=get_doi_url(dataset.identifier))
            if is_doi(dataset.identifier)
            else Url(url_id=remove_credentials(self.original_uri))
        )

        if dataset.description and not dataset.description.strip():
            dataset.description = None

        for creator in dataset.creators:
            if creator.affiliation == "":
                creator.affiliation = None

        self._provider_dataset_files = [
            ProviderDatasetFile(
                source=file.remote_url.geturl(),
                filename=Path(file.name).name,
                checksum="",
                filesize=file.content_size,
                filetype=file.file_format,
                path="",
            )
            for file in files
        ]

        self._provider_dataset = dataset
        return self._provider_dataset


class DataverseFileSerializer:
    """Dataverse record file."""

    def __init__(
        self,
        *,
        content_size=None,
        content_url=None,
        description=None,
        file_format=None,
        id=None,
        identifier=None,
        name=None,
        parent_url=None,
        type=None,
        encoding_format=None,
    ):
        self.content_size = content_size
        self.content_url = content_url
        self.description = description
        self.file_format = file_format
        self.id = id
        self.identifier = identifier
        self.name = name
        self.parent_url = parent_url
        self.type = type
        self.encoding_format = encoding_format

    @property
    def remote_url(self):
        """Get remote URL as ``urllib.ParseResult``."""
        if self.content_url is not None:
            return urllib.parse.urlparse(self.content_url)

        if self.identifier is None:
            return None

        doi = extract_doi(self.identifier)

        if doi is None:
            return None

        file_url = make_file_url("doi:" + doi, self.parent_url)

        return urllib.parse.urlparse(file_url)


class DataverseExporter(ExporterApi):
    """Dataverse export manager."""

    def __init__(self, *, dataset, server_url=None, dataverse_name=None, publish=False):
        super().__init__(dataset)
        self._access_token = None
        self._server_url = server_url
        self._dataverse_name = dataverse_name
        self._publish = publish

    def set_access_token(self, access_token):
        """Set access token."""
        self._access_token = access_token

    def get_access_token_url(self):
        """Endpoint for creation of access token."""
        return urllib.parse.urljoin(self._server_url, "/dataverseuser.xhtml?selectTab=apiTokenTab")

    def export(self, **kwargs):
        """Execute export process."""
        from renku.domain_model.dataset import get_file_path_in_dataset

        deposition = _DataverseDeposition(server_url=self._server_url, access_token=self._access_token)
        metadata = self._get_dataset_metadata()
        response = deposition.create_dataset(dataverse_name=self._dataverse_name, metadata=metadata)
        dataset_pid = response.json()["data"]["persistentId"]
        repository = project_context.repository

        with communication.progress("Uploading files ...", total=len(self.dataset.files)) as progressbar:
            for file in self.dataset.files:
                filepath = repository.copy_content_to_file(path=file.entity.path, checksum=file.entity.checksum)
                path_in_dataset = get_file_path_in_dataset(dataset=self.dataset, dataset_file=file)
                deposition.upload_file(full_path=filepath, path_in_dataset=path_in_dataset)
                progressbar.update()

        if self._publish:
            deposition.publish_dataset()

        return dataset_pid

    def _get_dataset_metadata(self):
        authors, contacts = self._get_creators()
        subject = self._get_subject()
        keywords = self._get_keywords()
        metadata_template = Template(DATASET_METADATA_TEMPLATE)
        metadata = metadata_template.substitute(
            name=_escape_json_string(self.dataset.name),
            authors=json.dumps(authors),
            contacts=json.dumps(contacts),
            description=_escape_json_string(self.dataset.description),
            subject=subject,
            keywords=json.dumps(keywords),
        )
        return json.loads(metadata)

    @staticmethod
    def _get_subject():
        text_prompt = "Subject of this dataset: \n\n"
        text_prompt += "\n".join(f"{s}\t[{i}]" for i, s in enumerate(DATAVERSE_SUBJECTS, start=1))
        text_prompt += "\n\nSubject"

        selection = communication.prompt(text_prompt, type=int, default=len(DATAVERSE_SUBJECTS)) or 0

        return DATAVERSE_SUBJECTS[selection - 1]

    def _get_creators(self):
        authors = []
        contacts = []

        for creator in self.dataset.creators:
            name = creator.name or ""
            affiliation = creator.affiliation or ""
            email = creator.email or ""

            author_template = Template(AUTHOR_METADATA_TEMPLATE)
            author = author_template.substitute(
                name=_escape_json_string(name), affiliation=_escape_json_string(affiliation)
            )
            authors.append(json.loads(author))

            contact_template = Template(CONTACT_METADATA_TEMPLATE)
            contact = contact_template.substitute(name=_escape_json_string(name), email=email)
            contacts.append(json.loads(contact))

        return authors, contacts

    def _get_keywords(self):
        keywords = []

        for keyword in self.dataset.keywords:
            keyword_template = Template(KEYWORDS_METADATA_TEMPLATE)
            keyword_str = keyword_template.substitute(keyword=_escape_json_string(keyword))
            keywords.append(json.loads(keyword_str))

        return keywords


class _DataverseDeposition:
    """Dataverse record for deposit."""

    def __init__(self, *, access_token, server_url, dataset_pid=None):
        self.access_token = access_token
        self.server_url = server_url
        self.dataset_pid = dataset_pid

    DATASET_CREATE_PATH = "dataverses/{dataverseName}/datasets"
    FILE_UPLOAD_PATH = "datasets/:persistentId/add"
    DATASET_PUBLISH_PATH = "datasets/:persistentId/actions/:publish"

    def create_dataset(self, dataverse_name, metadata):
        """Create a dataset in a given dataverse."""
        api_path = self.DATASET_CREATE_PATH.format(dataverseName=dataverse_name)
        url = self._make_url(api_path=api_path)

        response = self._post(url=url, json=metadata)
        self._check_response(response)

        self.dataset_pid = response.json()["data"]["persistentId"]

        return response

    def upload_file(self, full_path, path_in_dataset):
        """Upload a file to a previously-created dataset."""
        if self.dataset_pid is None:
            raise errors.ExportError("Dataset not created.")

        url = self._make_url(self.FILE_UPLOAD_PATH, persistentId=self.dataset_pid)

        params = {"directoryLabel": str(path_in_dataset.parent)}
        data = dict(jsonData=json.dumps(params))

        files = {"file": (path_in_dataset.name, open(full_path, "rb"))}

        response = self._post(url=url, data=data, files=files)
        self._check_response(response)

        return response

    def publish_dataset(self):
        """Publish a previously-created dataset."""
        if self.dataset_pid is None:
            raise errors.ExportError("Dataset not created.")

        url = self._make_url(self.DATASET_PUBLISH_PATH, persistentId=self.dataset_pid, type="major")

        response = self._post(url=url)
        self._check_response(response)

        return response

    def _make_url(self, api_path, **query_params):
        """Create URL for creating a dataset."""
        url_parts = urlparse.urlparse(self.server_url)
        path = posixpath.join(DATAVERSE_API_PATH, api_path)

        query_params_str = urllib.parse.urlencode(query_params)
        url_parts = url_parts._replace(path=path, query=query_params_str)
        return urllib.parse.urlunparse(url_parts)

    def _post(self, url, json=None, data=None, files=None):
        from renku.core.util import requests

        headers = {"X-Dataverse-key": self.access_token}
        try:
            return requests.post(url=url, json=json, data=data, files=files, headers=headers)
        except errors.RequestError as e:
            raise errors.ExportError("Cannot POST to remote server.") from e

    @staticmethod
    def _check_response(response):
        from renku.core.util import requests

        try:
            requests.check_response(response=response)
        except errors.RequestError:
            json_res = response.json()
            raise errors.ExportError(
                "HTTP {} - Cannot export dataset: {}".format(
                    response.status_code, json_res["message"] if "message" in json_res else json_res["status"]
                )
            )


def _escape_json_string(value):
    """Create a JSON-safe string."""
    if isinstance(value, str):
        return json.dumps(value)[1:-1]
    return value


def check_dataverse_uri(url):
    """Check if an URL points to a dataverse instance."""
    from renku.core.util import requests

    url_parts = list(urlparse.urlparse(url))
    url_parts[2] = posixpath.join(DATAVERSE_API_PATH, DATAVERSE_VERSION_API)

    url_parts[3:6] = [""] * 3
    version_url = urlparse.urlunparse(url_parts)

    response = requests.get(version_url)

    if response.status_code != 200:
        return False

    version_data = response.json()

    if "status" not in version_data or "data" not in version_data:
        return False

    version_info = version_data["data"]

    if "version" not in version_info or "build" not in version_info:
        return False

    return True


def check_dataverse_doi(doi):
    """Check if a DOI points to a dataverse dataset."""
    try:
        doi = DOIProvider(uri=doi).get_importer()
    except LookupError:
        return False

    return check_dataverse_uri(doi.uri)


def make_records_url(record_id, base_url):
    """Create URL to access record by ID."""
    url_parts = list(urlparse.urlparse(base_url))
    url_parts[2] = posixpath.join(DATAVERSE_API_PATH, DATAVERSE_METADATA_API)
    args_dict = {"exporter": DATAVERSE_EXPORTER, "persistentId": record_id}
    url_parts[4] = urllib.parse.urlencode(args_dict)
    return urllib.parse.urlunparse(url_parts)


def make_versions_url(record_id, base_url):
    """Create URL to access the versions of a record."""
    url_parts = list(urlparse.urlparse(base_url))
    url_parts[2] = posixpath.join(DATAVERSE_API_PATH, DATAVERSE_VERSIONS_API)
    args_dict = {"exporter": DATAVERSE_EXPORTER, "persistentId": record_id}
    url_parts[4] = urllib.parse.urlencode(args_dict)
    return urllib.parse.urlunparse(url_parts)


def make_file_url(file_id, base_url):
    """Create URL to access record by ID."""
    url_parts = list(urlparse.urlparse(base_url))
    url_parts[2] = posixpath.join(DATAVERSE_API_PATH, DATAVERSE_FILE_API)
    args_dict = {"persistentId": file_id}
    url_parts[4] = urllib.parse.urlencode(args_dict)
    return urllib.parse.urlunparse(url_parts)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Clone a Renku project.

Description
~~~~~~~~~~~

Clone a project and set up everything to use `renku` right away.

This command is preferred over using ``git clone`` because you don't need
to manually set up the required Git hooks, nor enable Git-LFS. It takes
care of that automatically.

Mind that, by default, ``renku clone`` pulls data from Git-LFS after cloning.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.clone:clone
   :prog: renku clone
   :nested: full


Handle special cases
~~~~~~~~~~~~~~~~~~~~

To clone private repositories with an HTTPS address, you first need to log into
a Renku deployment using the :ref:`cli-login` command. ``renku clone`` will use
the stored credentials when available.

To move a project to another Renku deployment you need to create a
new empty project in the target deployment and push both the
repository and Git-LFS objects to the new remote. Refer to Git documentation
for more details.

.. code-block:: console

    $ git lfs fetch --all
    $ git remote remove origin
    $ git remote add origin <new-repository-url>
    $ git push --mirror origin
"""

import click

import renku.ui.cli.utils.color as color


@click.command()
@click.option("--no-pull-data", is_flag=True, help="Do not pull data from Git-LFS.", default=False)
@click.argument("url")
@click.argument("path", required=False, default=None)
def clone(no_pull_data, url, path):
    """Clone a Renku repository."""
    from renku.command.clone import project_clone_command
    from renku.core.util.git import get_git_progress_instance

    click.echo(f"Cloning {url} ...")
    project_clone_command().build().execute(
        url=url, path=path, skip_smudge=no_pull_data, progress=get_git_progress_instance(), use_renku_credentials=True
    )
    click.secho("OK", fg=color.GREEN)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SSH utility functions."""

import textwrap
import urllib.parse
from pathlib import Path
from typing import NamedTuple, Optional, cast

from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

from renku.core import errors
from renku.core.session.utils import get_renku_url
from renku.core.util import communication
from renku.domain_model.project_context import project_context


class SSHKeyPair(NamedTuple):
    """A public/private key pair for SSH."""

    private_key: str
    public_key: str


def generate_ssh_keys() -> SSHKeyPair:
    """Generate an SSH key pair.

    Returns:
        Private Public key pair.
    """
    key = Ed25519PrivateKey.generate()

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.OpenSSH,
        crypto_serialization.NoEncryption(),
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH
    )

    return SSHKeyPair(private_key.decode("utf-8"), public_key.decode("utf-8"))


class SystemSSHConfig:
    """Class to manage system SSH config."""

    def __init__(self) -> None:
        """Initialize class and calculate paths."""
        self.ssh_root: Path = Path.home() / ".ssh"
        self.ssh_config: Path = self.ssh_root / "config"
        self.renku_ssh_root: Path = self.ssh_root / "renku"

        self.renku_ssh_root.mkdir(mode=0o700, exist_ok=True, parents=True)
        self.ssh_config.touch(mode=0o600, exist_ok=True)

        self.renku_host: Optional[str] = cast(Optional[str], urllib.parse.urlparse(get_renku_url()).hostname)

        if not self.renku_host:
            raise errors.AuthenticationError(
                "Please use `renku login` to log in to the remote deployment before setting up ssh."
            )

        self.jumphost_file = self.renku_ssh_root / f"99-{self.renku_host}-jumphost.conf"
        self.keyfile = self.renku_ssh_root / f"{self.renku_host}-key"
        self.public_keyfile = self.renku_ssh_root / f"{self.renku_host}-key.pub"

    @property
    def is_configured(self) -> bool:
        """Check if the system is already configured correctly."""
        return self.jumphost_file.exists() and self.keyfile.exists() and self.public_keyfile.exists()

    @property
    def public_key_string(self) -> Optional[str]:
        """Get the public key string, ready for authorized_keys."""
        try:
            key = self.public_keyfile.read_text()
            key = f"\n{key} {project_context.repository.get_user().name}"
            return key
        except FileNotFoundError:
            return None

    def is_session_configured(self, session_name: str) -> bool:
        """Check if a session is configured for SSH.

        Args:
            session_name(str): The name of the session.
        """
        if not project_context.ssh_authorized_keys_path.exists():
            return False

        session_commit = session_name.rsplit("-", 1)[-1]

        try:
            project_context.repository.get_commit(session_commit)
        except errors.GitCommitNotFoundError:
            return False

        try:
            authorized_keys = project_context.repository.get_content(
                project_context.ssh_authorized_keys_path, revision=session_commit
            )
        except errors.FileNotFound:
            return False

        if self.public_key_string and self.public_key_string in authorized_keys:
            return True
        return False

    def session_config_path(self, project_name: str, session_name: str) -> Path:
        """Get path to a session config.

        Args:
            project_name(str): The name of the project, without the owner name.
            session_name(str): The name of the session to setup a connection to.
        Returns:
            The path to the SSH connection file.
        """
        return self.renku_ssh_root / f"00-{project_name}-{session_name}.conf"

    def setup_session_keys(self) -> bool:
        """Add a users key to a project."""
        project_context.ssh_authorized_keys_path.parent.mkdir(parents=True, exist_ok=True)
        project_context.ssh_authorized_keys_path.touch(mode=0o600, exist_ok=True)

        if not self.public_key_string:
            raise errors.SSHNotSetupError()

        if self.public_key_string in project_context.ssh_authorized_keys_path.read_text():
            return False

        communication.info("Adding SSH public key to project.")
        with project_context.ssh_authorized_keys_path.open("at") as f:
            f.writelines(self.public_key_string)

        project_context.repository.add(project_context.ssh_authorized_keys_path)
        project_context.repository.commit("Add SSH public key.")
        communication.info(
            "Added public key. Changes need to be pushed and remote image built for changes to take effect."
        )
        return True

    def setup_session_config(self, project_name: str, session_name: str) -> str:
        """Setup local SSH config for connecting to a session.

        Args:
            project_name(str): The name of the project, without the owner name.
            session_name(str): The name of the session to setup a connection to.
        Returns:
            The name of the created SSH host config.
        """
        path = self.session_config_path(project_name, session_name)
        path.touch(mode=0o600, exist_ok=True)

        config_content = textwrap.dedent(
            f"""
            Host {session_name}
                HostName {session_name}
                RemoteCommand cd work/{project_name}/ || true && exec $SHELL --login
                RequestTTY yes
                ServerAliveInterval 15
                ServerAliveCountMax 3
                ProxyJump  jumphost-{self.renku_host}
                IdentityFile {self.keyfile}
                IdentityFile ~/.ssh/id_rsa
                IdentityFile ~/.ssh/id_ecdsa
                IdentityFile ~/.ssh/id_ecdsa_sk
                IdentityFile ~/.ssh/id_ed25519
                IdentityFile ~/.ssh/id_ed25519_sk
                IdentityFile ~/.ssh/id_dsa
                User jovyan
                StrictHostKeyChecking no
            """
        )

        path.write_text(config_content)

        return session_name



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migration utility functions."""

import os
from collections import defaultdict
from enum import IntFlag
from typing import Any, Dict, NamedTuple, Optional

from renku.core import errors
from renku.core.constant import RENKU_HOME
from renku.core.util import communication
from renku.infrastructure.repository import Repository, git_unicode_unescape


class MigrationType(IntFlag):
    """Type of migration that is being executed."""

    DATASETS = 1
    WORKFLOWS = 2
    STRUCTURAL = 4
    ALL = DATASETS | WORKFLOWS | STRUCTURAL


class MigrationOptions(NamedTuple):
    """Migration options."""

    strict: bool
    preserve_identifiers: bool
    type: MigrationType = MigrationType.ALL


class MigrationContext:
    """Context containing required migration information."""

    def __init__(self, strict: bool, type: MigrationType, preserve_identifiers: bool):
        self.options = MigrationOptions(strict=strict, type=type, preserve_identifiers=preserve_identifiers)
        self.dataset_migration_context: Optional[DatasetMigrationContext] = None
        self.cache: Optional[RepositoryCache] = None
        self.cwl_cache: Optional[Dict[str, Any]] = None
        self.plan_cache: Optional[Dict[str, Any]] = None


class DatasetMigrationContext:
    """The context to get path info in a specific revision."""

    def __init__(self, repository, revision):
        self.repository = repository
        self.revision = revision

    def exists(self, path) -> bool:
        """If a given path existed at the revision."""
        try:
            self.repository.run_git_command("cat-file", "-e", f"{self.revision}:{path}")
        except errors.GitCommandError:
            return False
        else:
            return True

    def is_dir(self, path) -> bool:
        """If a given path was a directory at the revision."""
        try:
            result = self.repository.run_git_command("cat-file", "-t", f"{self.revision}:{path}")
        except errors.GitCommandError:
            return False
        else:
            return "tree" in result

    def get_previous_commit(self, path):
        """Get previous commit of a given path starting from the revision."""
        return self.repository.get_previous_commit(path, revision=self.revision)


class RepositoryCache:
    """Cache for a git repository."""

    def __init__(self, repository: "Repository", cache, cwl_files_commits):
        self.repository = repository
        self.cache = cache
        self.cwl_files_commits = cwl_files_commits

    @classmethod
    def from_repository(cls, repository: "Repository"):
        """Return a cached repository."""
        cache = defaultdict(list)
        cwl_files_commits_map = {}

        for n, commit in enumerate(repository.iterate_commits(full_history=True), start=1):
            communication.echo(f"Caching commit {n}", end="\r")

            cwl_files = []
            for file in commit.get_changes():
                # Ignore deleted files
                if file.deleted:
                    continue

                path = file.b_path
                cache[path].append(commit)

                if path.startswith(f"{RENKU_HOME}/workflow/") and path.endswith(".cwl"):
                    cwl_files.append(os.path.realpath(repository.path / path))

            cls._update_cwl_files_and_commits(commit, cwl_files_commits_map, cwl_files)

        communication.echo(40 * " ", end="\r")

        return RepositoryCache(repository, cache, cwl_files_commits_map)

    @staticmethod
    def _update_cwl_files_and_commits(commit, cwl_files_commits_map, cwl_files):
        if len(cwl_files) != 1:
            return

        path = cwl_files[0]
        existing_commit = cwl_files_commits_map.get(path)

        if existing_commit is None:
            cwl_files_commits_map[path] = commit
        elif existing_commit.compare_to(commit) < 0:  # existing commit is older
            cwl_files_commits_map[path] = commit

    def find_previous_commit(self, path, revision="HEAD"):
        """Return a previous commit for a given path starting from 'revision'."""

        def find_from_repository(path, revision):
            try:
                return self.repository.get_previous_commit(path=path, revision=revision, full_history=True)
            except errors.GitCommitNotFoundError:
                communication.warn(f"Cannot find previous commit for {path} from {str(revision)}")
                return revision

        try:
            path = (self.repository.path / path).relative_to(self.repository.path)
        except ValueError:
            pass
        path = str(path)

        if revision == "HEAD":
            revision = self.repository.head.commit

        commits = self.cache.get(git_unicode_unescape(path))
        if not commits:
            return find_from_repository(path, revision)

        if revision in commits:
            return revision

        for commit in commits:
            if commit.compare_to(revision) <= 0:
                return commit

        # No commit was found
        return find_from_repository(path, revision)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku project gateway interface."""

from renku.core.interface.project_gateway import IProjectGateway
from renku.domain_model.project import Project
from renku.domain_model.project_context import project_context


class ProjectGateway(IProjectGateway):
    """Gateway for project database operations."""

    def get_project(self) -> Project:
        """Get project metadata."""
        try:
            return project_context.database["project"]  # type: ignore
        except KeyError as e:
            raise ValueError("Cannot get project") from e

    def update_project(self, project: Project):
        """Update project metadata."""
        from renku import __version__

        database = project_context.database

        try:
            if database["project"]:
                database.remove_root_object("project")
        except KeyError:
            pass

        project.agent_version = __version__

        database.add_root_object("project", project)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Models for providers."""

import dataclasses
import os
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, NamedTuple, Optional, Type

from humanize import naturalsize
from marshmallow import EXCLUDE

from renku.command.schema.dataset import DatasetSchema
from renku.domain_model.dataset import Dataset
from renku.infrastructure.immutable import DynamicProxy

if TYPE_CHECKING:
    from renku.core.dataset.providers.api import StorageProviderInterface
    from renku.domain_model.dataset import DatasetTag, RemoteEntity


class DatasetAddAction(Enum):
    """Types of action when adding a file to a dataset."""

    COPY = auto()
    MOVE = auto()
    SYMLINK = auto()
    NONE = auto()
    DOWNLOAD = auto()  # For URIs that are from a storage provider
    METADATA_ONLY = auto()  # For URIs that will be added to a dataset with a storage backend
    REMOTE_STORAGE = auto()  # For URIs that are from a remote storage provider


class DatasetUpdateAction(Enum):
    """Types of action when updating a file in a dataset."""

    UPDATE = auto()
    DELETE = auto()


@dataclasses.dataclass
class DatasetAddMetadata:
    """Metadata for a new file that will be added to a dataset."""

    entity_path: Path  # Entity path relative to the project's root
    url: str
    action: DatasetAddAction
    source: Path
    destination: Path
    provider: Optional["StorageProviderInterface"] = None
    based_on: Optional["RemoteEntity"] = None
    size: Optional[int] = None

    @property
    def has_action(self) -> bool:
        """Returns if file's action is not NONE."""
        return self.action != DatasetAddAction.NONE

    @property
    def metadata_only(self) -> bool:
        """Returns if file should be added to a remote storage."""
        return self.action == DatasetAddAction.METADATA_ONLY

    @property
    def from_cloud_storage(self) -> bool:
        """Returns if file is from a cloud storage."""
        return self.action == DatasetAddAction.REMOTE_STORAGE

    def get_absolute_commit_path(self, project_path: Path) -> str:
        """Return path of the file in the repository."""
        return os.path.join(project_path, self.entity_path)


@dataclasses.dataclass
class DatasetUpdateMetadata:
    """Metadata for updating dataset files."""

    entity: DynamicProxy
    action: DatasetUpdateAction


class ProviderParameter(NamedTuple):
    """Provider-specific parameters."""

    name: str
    default: Any = None
    flags: List[str] = []
    help: str = ""
    is_flag: bool = False
    multiple: bool = False
    type: Optional[Type] = None
    metavar: Optional[str] = None


class ProviderDataset(Dataset):
    """A Dataset that is imported from a provider."""

    def __init__(self, **kwargs):
        kwargs.setdefault("initial_identifier", "invalid-initial-id")
        super().__init__(**kwargs)
        self.dataset_files = []  # TODO Make this a property
        self._tag: Optional["DatasetTag"] = None

    @classmethod
    def from_jsonld(cls, data, schema_class=None) -> "ProviderDataset":
        """Create an instance from JSON-LD data."""
        assert isinstance(data, (dict, list)), f"Invalid data type: {data}"

        schema_class = schema_class or DatasetSchema
        self = schema_class(flattened=True).load(data)
        return self

    @classmethod
    def from_dataset(cls, dataset: "Dataset") -> "ProviderDataset":
        """Create an instance from a Dataset."""
        return ProviderDataset(
            annotations=dataset.annotations,
            creators=dataset.creators,
            dataset_files=[],
            date_created=dataset.date_created,
            date_published=dataset.date_published,
            date_removed=dataset.date_removed,
            derived_from=dataset.derived_from,
            description=dataset.description,
            id=dataset.id,
            identifier=dataset.identifier,
            images=dataset.images,
            in_language=dataset.in_language,
            initial_identifier=dataset.initial_identifier,
            keywords=dataset.keywords,
            license=dataset.license,
            slug=dataset.slug,
            project_id=dataset.project_id,
            same_as=dataset.same_as,
            name=dataset.name,
            version=dataset.version,
            storage=dataset.storage,
        )

    @property
    def files(self):
        """Return list of existing files."""
        raise NotImplementedError("ProviderDataset has no files.")

    @property
    def tag(self) -> Optional["DatasetTag"]:
        """Return dataset's tag."""
        return self._tag

    @tag.setter
    def tag(self, value):
        """Set dataset's tag."""
        self._tag = value


class ProviderDatasetFile:
    """Store metadata for dataset files that will be downloaded from a provider."""

    def __init__(
        self, source: Optional[str], filename: str, checksum: str, filesize: Optional[int], filetype: str, path: str
    ):
        self.checksum: str = checksum
        self.filename: str = filename
        self.filetype: str = filetype
        self.path: str = path
        self.filesize: Optional[int] = filesize
        self.filesize_str: Optional[str] = (
            naturalsize(filesize).upper().replace("BYTES", " B") if filesize is not None else None
        )
        self.source: Optional[str] = source


class ProviderDatasetSchema(DatasetSchema):
    """ProviderDataset schema."""

    class Meta:
        """Meta class."""

        model = ProviderDataset
        unknown = EXCLUDE



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for dataset tags list."""

from collections import OrderedDict
from typing import Callable, Dict


def tabular(tags):
    """Format dataset tags with a tabular output.

    Args:
        tags: Dataset tags.

    Returns:
        String of tags in tabular representation.
    """
    from renku.core.util.tabulate import tabulate

    return tabulate(
        tags,
        headers=OrderedDict(
            (
                ("date_created", "created"),
                ("name", None),
                ("description", None),
                ("dataset", None),
                ("dataset_id.value", "dataset id"),
            )
        ),
        # workaround for tabulate issue 181
        # https://bitbucket.org/astanin/python-tabulate/issues/181/disable_numparse-fails-on-empty-input
        disable_numparse=[1, 2, 4] if len(tags) > 0 else False,
    )


def jsonld(tags):
    """Format dataset tags as JSON-LD.

    Args:
        tags: Dataset tags.

    Returns:
        String of tags in JSON-LD representation.
    """
    from renku.command.format.json import dumps
    from renku.command.schema.dataset import DatasetTagSchema

    data = [DatasetTagSchema().dump(tag) for tag in tags]
    return dumps(data, indent=2)


DATASET_TAGS_FORMATS: Dict[str, Callable] = {
    "tabular": tabular,
    "json-ld": jsonld,
}
"""Valid formatting options."""
